from typing import Optional,List,Dict,Any,Union
import json

class ClusterSettings:
    def __init__(self, name:str, location:str, numNodes:int):
        self.name = name
        self.location = location 
        self.numNodes = numNodes   

    def json(self):
        return json.dumps(self.__dict__,ensure_ascii=False)        

    @staticmethod 
    def from_json(json_str:str):
        return ClusterSettings(**json.loads(json_str))       

class TableSettings:
    def __init__(self, database:str, table:Optional[str], schema:str, location:Optional[str], num_shards:int,status:str="open"):
        self.database = database
        self.table = table
        self.schema = schema
        self.location = location
        self.num_shards = num_shards
        self.status = status

    def json(self):
        return json.dumps(self.__dict__,ensure_ascii=False)

    @staticmethod 
    def from_json(json_str:str):
        return TableSettings(**json.loads(json_str))    


class EnvSettings:
    def __init__(self, javaHome:str, path:str):
        self.javaHome = javaHome
        self.path = path   

    def json(self):
        return json.dumps(self.__dict__,ensure_ascii=False) 

    @staticmethod
    def from_json(json_str:str):
        return EnvSettings(**json.loads(json_str)) 


class ResourceRequirement:
    def __init__(self, name:str, resourceQuantity:float):
        self.name = name
        self.resourceQuantity = resourceQuantity   

    def json(self):
        return json.dumps(self.__dict__,ensure_ascii=False)
    
    @staticmethod
    def from_json(json_str:str):
        return ResourceRequirement(**json.loads(json_str))
    


class ResourceRequirementSettings:
    def __init__(self, resourceRequirements: List[ResourceRequirement]):
        self.resourceRequirements = resourceRequirements

    def json(self):
        return json.dumps({"resourceRequirements":[item.__dict__ for item in self.resourceRequirements]},ensure_ascii=False)           
    
    @staticmethod
    def from_json(json_str:str):
        s = json.loads(json_str)
        return ResourceRequirementSettings([ResourceRequirement(**s["resourceRequirements"]) ])


class JVMSettings:
    def __init__(self, options:list[str]):
        self.options = options   

    def json(self):
        return json.dumps(self.__dict__,ensure_ascii=False) 

    @staticmethod
    def from_json(json_str:str):
        return JVMSettings(**json.loads(json_str))        


class SearchQuery:
    '''
    filters: List[Dict[str,Any]] = {"and":[{"field":"name","value":"张三"}]}
    filters: List[Dict[str,Any]] = {"or":[{"field":"name","value":"张三"},{"field":"name","value":"李四"}]}
    filters: List[Dict[str,Any]] = {"or":[{"field":"name","value":"张三"},{"and":[{"field":"name","value":"李四"},{"field":"age","min":10,"max":20}]}]}}]}    
    '''
    def __init__(self,database:str,
                 table:str,                  
                 keyword:Optional[str], fields:list[str], 
                 vector:list[float], vectorField:Optional[str], filters:Dict[str,Any]={},
                 sorts: List[Dict[str,str]]=[],limit:int=10):
        self.database = database
        self.table = table
        self.filters = filters
        self.sorts = sorts
        self.keyword = keyword
        self.fields = fields
        self.vector = vector
        self.vectorField = vectorField
        self.limit = limit

    def json(self):
        return json.dumps(self.__dict__,ensure_ascii=False) 

    @staticmethod
    def from_json(json_str:str):
        return SearchQuery(**json.loads(json_str)) 


# class EqualFilter:
#     def __init__(self, field:str, value:Any):
#         self.field = field
#         self.value = value

#     def json(self):
#         return json.dumps(self.__dict__,ensure_ascii=False) 

#     @staticmethod
#     def from_json(json_str:str):
#         return EqualFilter(**json.loads(json_str))

# class RangeFilter:
#     def __init__(self, field:str, min:Any, max:Any):
#         self.field = field
#         self.min = min
#         self.max = max

#     def json(self):
#         return json.dumps(self.__dict__,ensure_ascii=False) 

#     @staticmethod
#     def from_json(json_str:str):
#         return RangeFilter(**json.loads(json_str))

# class OrRelation:
#     def __init__(self, filters:Union[EqualFilter,RangeFilter],parent:Union["OrRelation","AndRelation","QueryBuilder"]):
#         self.filters = filters

#     def json(self):
#         return json.dumps(self.__dict__,ensure_ascii=False) 

#     @staticmethod
#     def from_json(json_str:str):
#         return OrRelation(**json.loads(json_str)) 
    
#     def and_relation(self):
#         return AndRelation(self)        
    
#     def equal_filter(self,field:str,value:Any):
#         self.filters.append(EqualFilter(field,value))
#         return self
    
#     def range_filter(self,field:str,min:Any,max:Any):
#         self.filters.append(RangeFilter(field,min,max))
#         return self
    
#     def end(self):
#         return self.parent
    
#     def or_relation(self):
#         return OrRelation(self)

# class AndRelation:
#     def __init__(self, filters:Union[EqualFilter,RangeFilter],parent:Union[OrRelation,"AndRelation","QueryBuilder"]):
#         self.filters = filters
#         self.parent = parent        

#     def json(self):
#         return json.dumps(self.__dict__,ensure_ascii=False) 

#     @staticmethod
#     def from_json(json_str:str):
#         return AndRelation(**json.loads(json_str))
    
#     def equal_filter(self,field:str,value:Any):
#         self.filters.append(EqualFilter(field,value))
#         return self
    
#     def range_filter(self,field:str,min:Any,max:Any):
#         self.filters.append(RangeFilter(field,min,max))
#         return self
    
#     def end(self):
#         return self.parent
    
#     def or_relation(self):
#         return OrRelation(self)
            

# class FilterBuilder:
#     def __init__(self,query_builder:"QueryBuilder"):
#         self._filter = []
#         self.query_builder = query_builder

#     def end(self):
#         return self.query_builder
    
#     def or_relation(self):
#         return OrRelation(self)
    


# class QueryBuilder:
#     def __init__(self):
#         self._database = ""
#         self._table = ""
#         self._filter = []
#         self._keyword = ""
#         self._fields = []
#         self._vector = []
#         self._vectorField = ""
#         self._limit = 10
#         self.filter_builder = FilterBuilder(self)

#     def database(self,database:str):
#         self._database = database
#         return self

#     def table(self,table:str):
#         self._table = table
#         return self

#     def filter(self):
#         return self.filter_builder


      

        
    



import asyncio
from collections import defaultdict
import copy
import inspect
import time
import logging
import uuid
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union,Generator
import ray
import concurrent
from ray.util.client.common import ClientActorHandle, ClientObjectRef
from byzerllm.utils.client import message_utils
from .agent import Agent
from ...utils.retrieval import ByzerRetrieval
from ...utils.client import ByzerLLM,default_chat_wrapper,LLMResponse
from . import get_agent_name,run_agent_func, ChatResponse
from .store import MessageStore,Message as ChatStoreMessage
from .store.stores import Stores

try:
    from termcolor import colored
except ImportError:

    def colored(x, *args, **kwargs):
        return x


logger = logging.getLogger(__name__)


class ConversableAgent(Agent):
    
    def __init__(
        self,
        name: str,
        llm: ByzerLLM,
        retrieval: ByzerRetrieval,                
        system_message: Optional[str] = "You are a helpful AI Assistant.",
        is_termination_msg: Optional[Callable[[Dict], bool]] = None,
        max_consecutive_auto_reply: Optional[int] = None,
        human_input_mode: Optional[str] = "TERMINATE",
        function_map: Optional[Dict[str, Callable]] = None,
        code_execution_config: Optional[Union[Dict, bool]] = None,        
        default_auto_reply: Optional[Union[str, Dict, None]] = "",
        chat_wrapper:Optional[Callable[[ByzerLLM,Optional[List[Dict]],Dict],List[LLMResponse]]] = None,
        description:str = "ConversableAgent",
        message_store: Optional[Union[str,ClientActorHandle,MessageStore]] = None,
        group_name: Optional[str] = None,
        
    ):
        super().__init__(name)
        
        self.llm = llm
        self.retrieval = retrieval   
        self.chat_wrapper = chat_wrapper
     
        self.message_store = None
        
        if message_store is not None:
            self.message_store = Stores(message_store)

        self.group_name = group_name    

        self._messages = defaultdict(list)
        self._system_message = [{"content": system_message, "role": "system"}]
        
        self._is_termination_msg = (
            is_termination_msg if is_termination_msg is not None else (
                lambda x:x.get("content", "").rstrip().endswith("TERMINATE") or x.get("metadata",{}).get("TERMINATE",False)  or x.get("content", "").rstrip().endswith("终止") or x.get("metadata",{}).get("终止",False) 
                                                                       )
        )
        self.human_input_mode = human_input_mode
        self._max_consecutive_auto_reply = (
            max_consecutive_auto_reply if max_consecutive_auto_reply is not None else -1
        )
        self._code_execution_config = {} if code_execution_config is None else code_execution_config
        self._consecutive_auto_reply_counter = defaultdict(int)
        self._max_consecutive_auto_reply_dict = defaultdict(self.max_consecutive_auto_reply)
        self._function_map = {} if function_map is None else function_map
        self._default_auto_reply = default_auto_reply
        self._reply_func_list = []
        self.reply_at_receive = defaultdict(lambda: True)
        self._agent_description = description
        
        self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.generate_llm_reply)           
        self.auto_register_reply()
        self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.check_termination_and_human_reply)         
        
        self.stream_replies = {} 
        self.error_count = {}
        
    def get_then_increment_error_count(self,agent:Union[ClientActorHandle,Agent,str]):
        agent_name = get_agent_name(agent)
        if agent_name not in self.error_count:
            self.error_count[agent_name] = 0
        v = self.error_count[agent_name]    
        self.error_count[agent_name] += 1
        return v    

    def auto_register_reply(self):        
        for _, attr in inspect.getmembers(self, predicate=inspect.ismethod):            
            if hasattr(attr, '_is_reply'):
                self.register_reply([Agent, ClientActorHandle,str], attr) 
                
    def stream_reply(self,response_gen,**kwargs):                
        id = str(uuid.uuid4())        
        def gen(): 
            t = ""           
            for response in response_gen:
                t += response
                yield (t,None)

        self._put_stream_reply(id,gen())
        return True, {
            "content":id,
            "metadata":{"agent":self.name,"TERMINATE":True,"stream":True,"stream_id":id,**kwargs}
        }                   

    def _put_stream_reply(self,id:str,reply:Generator): 
        self.stream_replies[id] = reply

    def _stream_get_message_from_self(self,id:str):
        for item in self.stream_replies.get(id,[]):
            yield item      

    def stream_get_message(self,agent:Union[ClientActorHandle,Agent,str],id:str):
        '''
        get stream reply from agent
        '''
        if isinstance(agent,Agent):
            return agent._stream_get_message_from_self(id)
        elif isinstance(agent,str):
            t = ray.get_actor(agent) 
            return t._stream_get_message_from_self.remote(id)
        else:
            return agent._stream_get_message_from_self.remote(id)
    
    def get_name(self) -> str:
        return self._name    

    def get_function_map(self):
        """Get the function map."""
        return self._function_map
    
    def get_agent_description(self):
        return self._agent_description
    
    def update_agent_description(self,description:str):
        self._agent_description = description                       
    
    def register_reply(
        self,
        trigger: Union[Type[Agent], str, Agent, Callable[[Agent], bool], List],
        reply_func: Callable,
        position: Optional[int] = 0,
        config: Optional[Any] = None,
        reset_config: Optional[Callable] = None,
    ):
        """Register a reply function.

        The reply function will be called when the trigger matches the sender.
        The function registered later will be checked earlier by default.
        To change the order, set the position to a positive integer.

        Args:
            trigger (Agent class, str, Agent instance, callable, or list): the trigger.
                - If a class is provided, the reply function will be called when the sender is an instance of the class.
                - If a string is provided, the reply function will be called when the sender's name matches the string.
                - If an agent instance is provided, the reply function will be called when the sender is the agent instance.
                - If a callable is provided, the reply function will be called when the callable returns True.
                - If a list is provided, the reply function will be called when any of the triggers in the list is activated.
                - If None is provided, the reply function will be called only when the sender is None.
                Note: Be sure to register `None` as a trigger if you would like to trigger an auto-reply function with non-empty messages and `sender=None`.
            reply_func (Callable): the reply function.
                The function takes a recipient agent, a list of messages, a sender agent and a config as input and returns a reply message.
        ```python
        def reply_func(
            recipient: ConversableAgent,
            messages: Optional[List[Dict]] = None,
            sender: Optional[Agent] = None,
            config: Optional[Any] = None,
        ) -> Union[str, Dict, None]:
        ```
            position (int): the position of the reply function in the reply function list.
                The function registered later will be checked earlier by default.
                To change the order, set the position to a positive integer.
            config (Any): the config to be passed to the reply function.
                When an agent is reset, the config will be reset to the original value.
            reset_config (Callable): the function to reset the config.
                The function returns None. Signature: ```def reset_config(config: Any)```
        """        
        self._reply_func_list.insert(
            position,
            {
                "trigger": trigger,
                "reply_func": reply_func,
                "config": copy.copy(config),
                "init_config": config,
                "reset_config": reset_config,
            },
        )

    @property
    def system_message(self):
        """Return the system message."""
        return self._system_message[0]["content"]
    
    def get_system_message(self):
        return self.system_message

    def update_system_message(self, system_message: str):
        """Update the system message.

        Args:
            system_message (str): system message for the ChatCompletion inference.
        """
        self._system_message[0]["content"] = system_message  

    def update_max_consecutive_auto_reply(self, value: int, sender: Optional[Union[Agent,ClientActorHandle,str]] = None):
        """Update the maximum number of consecutive auto replies.

        Args:
            value (int): the maximum number of consecutive auto replies.
            sender (Agent): when the sender is provided, only update the max_consecutive_auto_reply for that sender.
        """
        if sender is None:
            self._max_consecutive_auto_reply = value
            for k in self._max_consecutive_auto_reply_dict:
                self._max_consecutive_auto_reply_dict[k] = value
        else:
            self._max_consecutive_auto_reply_dict[get_agent_name(sender)] = value

    def max_consecutive_auto_reply(self, sender: Optional[Union[Agent,ClientActorHandle]] = None) -> int:
        """The maximum number of consecutive auto replies."""
        return self._max_consecutive_auto_reply if sender is None else self._max_consecutive_auto_reply_dict[get_agent_name(sender)]

    @property
    def chat_messages(self) -> Dict[Agent, List[Dict]]:
        """A dictionary of conversations from agent to list of messages."""
        return self._messages   

    def get_chat_messages(self):
        return self.chat_messages 
    
    def last_message(self, agent: Optional[Union[Agent,ClientActorHandle,str]] = None) -> Dict:
        """The last message exchanged with the agent.

        Args:
            agent (Agent): The agent in the conversation.
                If None and more than one agent's conversations are found, an error will be raised.
                If None and only one conversation is found, the last message of the only conversation will be returned.

        Returns:
            The last message exchanged with the agent.
        """
        if agent is None:
            n_conversations = len(self._messages)
            if n_conversations == 0:
                return None
            if n_conversations == 1:
                for conversation in self._messages.values():
                    return conversation[-1]
            raise ValueError("More than one conversation is found. Please specify the sender to get the last message.")
        if get_agent_name(agent) not in self._messages.keys():
            raise KeyError(
                f"The agent '{get_agent_name(agent)}' is not present in any conversation. No history available for this agent."
            )
        return self._messages[get_agent_name(agent)][-1]
    
    @staticmethod
    def _message_to_dict(message: Union[Dict, str]):
        """Convert a message to a dictionary.

        The message can be a string or a dictionary. The string will be put in the "content" field of the new dictionary.
        """
        if isinstance(message, str):
            return {"content": message}
        elif isinstance(message, dict):
            return message
        else:
            return dict(message)

    def _append_message(self, message: Union[Dict, str], role, conversation_id: Union[ClientActorHandle,Agent,str]) -> bool:
        """Append a message to the ChatCompletion conversation.

        If the message received is a string, it will be put in the "content" field of the new dictionary.
        If the message received is a dictionary but does not have any of the two fields "content" or "function_call",
            this message is not a valid ChatCompletion message.
        If only "function_call" is provided, "content" will be set to None if not provided, and the role of the message will be forced "assistant".

        Args:
            message (dict or str): message to be appended to the ChatCompletion conversation.
            role (str): role of the message, can be "assistant" or "function".
            conversation_id (Agent): id of the conversation, should be the recipient or sender.

        Returns:
            bool: whether the message is appended to the ChatCompletion conversation.
        """
        raw_message = message
        
        if isinstance(message, ChatResponse):
            message = {"content":raw_message.output,"metadata":{"raw_message":raw_message}}

        message = self._message_to_dict(message)
        # create oai message to be appended to the oai conversation that can be passed to oai directly.
        oai_message = {k: message[k] for k in ("content", "function_call", "name", "context","metadata") if k in message}
        if "content" not in oai_message:
            if "function_call" in oai_message:
                oai_message["content"] = None  # if only function_call is provided, content will be set to None.
            else:
                return False

        oai_message["role"] = "function" if message.get("role") == "function" else role
        if "function_call" in oai_message:
            oai_message["role"] = "assistant"  # only messages with role 'assistant' can have a function call.
            oai_message["function_call"] = dict(oai_message["function_call"])
        self._messages[get_agent_name(conversation_id)].append(oai_message)
        return True    


    def reset(self):
        """Reset the agent."""
        self.clear_history()
        self.reset_consecutive_auto_reply_counter()
        self.stop_reply_at_receive()
        for reply_func_tuple in self._reply_func_list:
            if reply_func_tuple["reset_config"] is not None:
                reply_func_tuple["reset_config"](reply_func_tuple["config"])
            else:
                reply_func_tuple["config"] = copy.copy(reply_func_tuple["init_config"])

    def stop_reply_at_receive(self, sender: Optional[Union[ClientActorHandle,Agent,str]]  = None):
        """Reset the reply_at_receive of the sender."""
        if sender is None:
            self.reply_at_receive.clear()
        else:
            self.reply_at_receive[get_agent_name(sender)] = False

    def reset_consecutive_auto_reply_counter(self, sender: Optional[Union[ClientActorHandle,Agent,str]]  = None):
        """Reset the consecutive_auto_reply_counter of the sender."""
        if sender is None:
            self._consecutive_auto_reply_counter.clear()
        else:
            self._consecutive_auto_reply_counter[get_agent_name(sender)] = 0

    def clear_history(self, agent: Optional[Union[ClientActorHandle,Agent,str]] = None):
        """Clear the chat history of the agent.

        Args:
            agent: the agent with whom the chat history to clear. If None, clear the chat history with all agents.
        """
        if agent is None:
            self._messages.clear()
        else:
            self._messages[get_agent_name(agent)].clear()

    
    def set_reply_at_receive(self, sender: Optional[Union[ClientActorHandle,Agent,str]] = None, value: bool = True):
        self.reply_at_receive[get_agent_name(sender)] = value 

    def get_reply_at_receive(self, sender: Optional[Union[ClientActorHandle,Agent,str]] = None):
        return self.reply_at_receive[get_agent_name(sender)] if sender is not None else self.reply_at_receive            

    def _prepare_chat(self, recipient, clear_history):            
            self.reset_consecutive_auto_reply_counter(recipient)

            # recipient.reset_consecutive_auto_reply_counter(self)
            run_agent_func(recipient, "reset_consecutive_auto_reply_counter", self)
                        
            # recipient.reply_at_receive[self] = True 
            run_agent_func(recipient, "set_reply_at_receive", self, True)
            self.reply_at_receive[get_agent_name(recipient)] = True
            if clear_history:
                self.clear_history(recipient)
                # recipient.clear_history(self)
                run_agent_func(recipient, "clear_history", self)

    def generate_init_message(self, **context) -> Union[str, Dict]:
        """Generate the initial message for the agent.

        Override this function to customize the initial message based on user's request.
        If not overriden, "message" needs to be provided in the context.
        """
        return context["message"]            

    def initiate_chat(
        self,
        recipient: Union[ClientActorHandle,Agent,str],
        clear_history: Optional[bool] = True,
        silent: Optional[bool] = False,
        **context,
    ): 
        self._prepare_chat(recipient, clear_history)
        self.send(self.generate_init_message(**context), recipient, silent=silent)
    
    def send(
        self,
        message: Union[Dict, str],
        recipient: Union[ClientActorHandle,Agent,str], #"Agent"
        request_reply: Optional[bool] = None,
        silent: Optional[bool] = False,
        async_send: Optional[bool] = False,
    ) -> bool:
        valid = self._append_message(message, "assistant", recipient)
        if valid:
            if isinstance(recipient, Agent):
                if async_send:
                    with concurrent.ThreadPoolExecutor() as executor:
                        executor.submit(recipient.receive, message, self, request_reply, silent)                    
                else:
                    recipient.receive(message, self, request_reply, silent)
            elif isinstance(recipient, str):                
                t = ray.get_actor(recipient) 
                if async_send:
                    t.receive.remote(message, self.get_name(), request_reply, silent)   
                else:            
                    ray.get(t.receive.remote(message, self.get_name(), request_reply, silent))
            else:
                if async_send:
                    recipient.receive.remote(message, self.get_name(), request_reply, silent)
                else:
                    ray.get(recipient.receive.remote(message, self.get_name(), request_reply, silent))    

            return True        
        else:
            raise ValueError(
                "Message can't be converted into a valid ChatCompletion message. Either content or function_call must be provided."
            ) 
        
    
    def _process_received_message(self, message, sender, silent):
            raw_message = message
            if isinstance(message, ChatResponse):                
                message = {"content":raw_message.output,"metadata":{"raw_message":raw_message}}
            message = self._message_to_dict(message)
            # When the agent receives a message, the role of the message is "user". (If 'role' exists and is 'function', it will remain unchanged.)
            valid = self._append_message(message, "user", sender)
            if not valid:
                raise ValueError(
                    "Received message can't be converted into a valid ChatCompletion message. Either content or function_call must be provided."
                )
            
            if not silent:  
                if self.message_store is not None:
                    self.message_store.put(ChatStoreMessage(
                        id=self.group_name if self.group_name is not None else "default",
                        m=message,
                        sender=get_agent_name(sender),
                        receiver=self.get_name(),
                        timestamp=time.monotonic()
                    ))
                                  
                print(colored(get_agent_name(sender), "yellow"), "(to", f"{self.name}):\n", flush=True)
                print(colored(f"{message['content']}", "green"), flush=True)
                print("\n", "-" * 80, flush=True, sep="")

    def receive(
        self,
        message: Union[Dict, str,ChatResponse],
        sender: Union[ClientActorHandle,Agent,str], #"Agent"
        request_reply: Optional[bool] = None,
        silent: Optional[bool] = False,
    ):        
        self._process_received_message(message, sender, silent)

        if request_reply is False or request_reply is None and self.reply_at_receive[get_agent_name(sender)] is False:            
            return
        reply = self.generate_reply(raw_message=message, messages=self.chat_messages[get_agent_name(sender)], sender=sender)
                
        if reply is not None:                                    
            self.send(reply, sender, silent=silent)

    def generate_reply(
        self,
        raw_message: Optional[Union[Dict,str,ChatResponse]] = None,
        messages: Optional[List[Dict]] = None,
        sender: Optional[Union[ClientActorHandle,Agent,str]] = None,
        exclude: Optional[List[Callable]] = None,
    ) -> Union[str, Dict, None,ChatResponse]:
        if all((messages is None, sender is None)):
            error_msg = f"Either {messages=} or {sender=} must be provided."
            logger.error(error_msg)
            raise AssertionError(error_msg)
        
        if messages is None:
            messages = self._messages[get_agent_name(sender)]                

        for reply_func_tuple in self._reply_func_list:
            reply_func = reply_func_tuple["reply_func"]
            if exclude and reply_func in exclude:
                continue
            if asyncio.coroutines.iscoroutinefunction(reply_func):
                continue
            # print(f'''{self.get_name()} generating reply for {get_agent_name(sender)} from {reply_func.__name__}''',flush=True)
            final, reply = reply_func(self, raw_message=raw_message, messages=messages, sender=sender, config=reply_func_tuple["config"])
            if final:                
                return reply
                         
        return self._default_auto_reply 
        
    def generate_llm_reply(
            self,
            raw_message: Optional[Union[Dict,str,ChatResponse]] = None,
            messages: Optional[List[Dict]] = None,
            sender: Optional[Union[ClientActorHandle,Agent,str]] = None,
            config: Optional[Any] = None,
        ) -> Tuple[bool, Union[str, Dict, None]]:
            """Generate a reply using autogen.oai."""            
            if self.llm is None:
                return False, None
            if messages is None:
                messages = self._messages[get_agent_name(sender)]

            # TODO: #1143 handle token limit exceeded error  
            # padding the messages to user/assistant pair
            # [{'content': '', 'role': 'assistant'},{'content': '', 'role': 'user'}, {'content': '', 'role': 'assistant'},{'content': '', 'role': 'assistant'}]    
            # should be converted to
            # [{'content': '', 'role': 'user'},{'content': '', 'role': 'assistant'},{'content': '', 'role': 'user'}, {'content': '', 'role': 'assistant'},{'content': '', 'role': 'user'},{'content': '', 'role': 'assistant'},{'content': '', 'role': 'user'}]                
            temp_messages = message_utils.padding_messages_merge(messages)

            if self.chat_wrapper is None:
                response = self.llm.chat_oai(conversations=self._system_message + temp_messages,llm_config={
                    "temperature":0.1,"top_p":0.95
                })
                return True, response[0].output
                        
            response = self.chat_wrapper(self.llm,self._system_message + temp_messages)
            return True, response[0].output    

    def get_human_input(self, prompt: str) -> str:
            """Get human input.

            Override this method to customize the way to get human input.

            Args:
                prompt (str): prompt for the human input.

            Returns:
                str: human input.
            """
            reply = input(prompt)
            return reply
    
    def check_termination_and_human_reply(
        self,
        raw_message: Optional[Union[Dict,str,ChatResponse]] = None,
        messages: Optional[List[Dict]] = None,
        sender: Optional[Union[ClientActorHandle,Agent,str]] = None,
        config: Optional[Any] = None,
    ) -> Tuple[bool, Union[str, Dict, None]]:        

        """Check if the conversation should be terminated, and if human reply is provided."""
        if config is None:
            config = self
        if messages is None:
            messages = self._messages[get_agent_name(sender)]
        message = messages[-1]
        reply = ""
        no_human_input_msg = ""
        if self.human_input_mode == "ALWAYS":
            reply = self.get_human_input(
                f"Provide feedback to {get_agent_name(sender)}. Press enter to skip and use auto-reply, or type 'exit' to end the conversation: "
            )
            no_human_input_msg = "NO HUMAN INPUT RECEIVED." if not reply else ""
            # if the human input is empty, and the message is a termination message, then we will terminate the conversation
            reply = reply if reply or not self._is_termination_msg(message) else "exit"
        else:            
            if self._max_consecutive_auto_reply_dict[get_agent_name(sender)] != -1 and self._consecutive_auto_reply_counter[get_agent_name(sender)] >= self._max_consecutive_auto_reply_dict[get_agent_name(sender)]:
                if self.human_input_mode == "NEVER":
                    reply = "exit"
                else:
                    # self.human_input_mode == "TERMINATE":
                    terminate = self._is_termination_msg(message)
                    reply = self.get_human_input(
                        f"Please give feedback to {get_agent_name(sender)}. Press enter or type 'exit' to stop the conversation: "
                        if terminate
                        else f"Please give feedback to {get_agent_name(sender)}. Press enter to skip and use auto-reply, or type 'exit' to stop the conversation: "
                    )
                    no_human_input_msg = "NO HUMAN INPUT RECEIVED." if not reply else ""
                    # if the human input is empty, and the message is a termination message, then we will terminate the conversation
                    reply = reply if reply or not terminate else "exit"
            elif self._is_termination_msg(message):
                if self.human_input_mode == "NEVER":
                    reply = "exit"
                else:
                    # self.human_input_mode == "TERMINATE":
                    reply = self.get_human_input(
                        f"Please give feedback to {sender.name}. Press enter or type 'exit' to stop the conversation: "
                    )
                    no_human_input_msg = "NO HUMAN INPUT RECEIVED." if not reply else ""
                    # if the human input is empty, and the message is a termination message, then we will terminate the conversation
                    reply = reply or "exit"

        # print the no_human_input_msg
        if no_human_input_msg:
            print(colored(f"\n>>>>>>>> {no_human_input_msg}", "red"), flush=True)

        # stop the conversation
        if reply == "exit":
            # reset the consecutive_auto_reply_counter
            self._consecutive_auto_reply_counter[get_agent_name(sender)] = 0
            return True, None

        # send the human reply
        if reply or self._max_consecutive_auto_reply_dict[get_agent_name(sender)] == 0:
            # reset the consecutive_auto_reply_counter
            self._consecutive_auto_reply_counter[get_agent_name(sender)] = 0
            return True, reply

        # increment the consecutive_auto_reply_counter        
        self._consecutive_auto_reply_counter[get_agent_name(sender)] += 1
        if self.human_input_mode != "NEVER":
            print(colored("\n>>>>>>>> USING AUTO REPLY...", "red"), flush=True)

        return False, None

    
                
    


"""
Utility functions to help with the tools api commonly used in LLMs.
"""

import inspect
from typing import Callable, Optional, Dict, List, Any, get_type_hints, get_origin

try:
    # For Python 3.9 and later
    from typing import _AnnotatedAlias  # type: ignore
except ImportError:
    # For Python versions below 3.9
    from typing_extensions import _AnnotatedAlias  # type: ignore

_type_map = {
    str: "string",
    int: "integer",
    float: "number",
    bool: "boolean",
    list: "array",
}


def _original_type_backward_compatibility(param_type):
    """
    This function is for backward compatibility with Python 3.8 and below.
    It returns the original type of the parameter, which is the same as the
    param_type if the Python version is 3.9 and above, but is the __origin__ of
    the param_type if the Python version is 3.8 and below.
    """
    try:
        return param_type.__origin__
    except AttributeError:
        return param_type


def _get_type_spec(param_type, param_annotation, default_value):
    """
    Gets the type spec in json format for a specific parameter type.

    The param type must be one of the following: str, int, float, bool, list, or typing.List. There are two ways to annotate a parameter: one is via Annotated[type, description], and one is via the default value as the description.

    The description depends on the parameter:
    - For list types, it should be a list of tuples, where each tuple is of length 3, and the first element is a string as the name, the second element is the type of the field, and the third element is the description of the field.
    - For string types, it should either be a string as the description, or a tuple of length 2, where the first element is the string description, and the second element is a list of strings representing the enum.
    - For all other types, it should be a string as the description.

    For example, for an int field, you can do
        def foo(int_param: Annotated[int, "this is an int description"])
    or
        def foo(int_param: int = "this is an int description")
    (note that the default value must be a string in this case, which is a bit hacky, but it works.
    It is recommended that you use Annotated whenever possible.)
    """
    if get_origin(param_type) in (list, List):
        param_type = list
    try:
        type_name = _type_map[param_type]
    except KeyError:
        raise TypeError(f"Type {param_type} is not supported by the api.")
    # We will first prefer the annotation, then the default value, to find the
    # description of the parameter.
    if isinstance(param_annotation, _AnnotatedAlias):
        description = param_annotation.__metadata__
        description = (
            description[0]
            if isinstance(description, tuple) and len(description) == 1
            else description
        )
    elif default_value is not inspect.Parameter.empty:
        description = default_value
    else:
        raise ValueError("Either param_annotation or default_value must be provided.")

    if param_type == list:
        if not isinstance(description, list):
            raise TypeError(
                f"For list type {param_type}(aka {type_name}), value must be a list"
                " containing the description of the field."
            )
        array_description = {"type": "object", "properties": {}}
        for i, v in enumerate(description):
            if len(v) == 2:
                sub_param_name = _original_type_backward_compatibility(v[0])
                sub_param_annotation = v[1]
                sub_param_type = v[1].__origin__
                sub_param_default_value = inspect.Parameter.empty
            elif len(v) == 3:
                sub_param_name = _original_type_backward_compatibility(v[0])
                sub_param_annotation = v[1]
                sub_param_type = v[1]
                sub_param_default_value = v[2]
            else:
                raise TypeError(
                    "For array type, each element of the list must be a tuple of"
                    " length 2 or 3, where the first element is a string, the second"
                    " element is the Annotated type (if len==2) or raw type (if"
                    " len==3) of the field, and the third element (if len==3) is the"
                    f" description of the field. Got {v} (index {i})"
                )
            try:
                type_spec = _get_type_spec(
                    sub_param_type, sub_param_annotation, sub_param_default_value
                )
            except Exception as e:
                raise TypeError(
                    f"Error when processing the {i}th element of the list {v}. Source"
                    f" exception: {e}"
                )
            array_description["properties"][sub_param_name] = type_spec
        return {"type": "array", "items": array_description}
    elif param_type == str:
        if isinstance(description, str):
            # simple string type
            return {"type": type_name, "description": description}
        elif (
            len(description) == 2
            and isinstance(description[0], str)
            and isinstance(description[1], list)
        ):
            # string type with enum
            if not all(isinstance(v, str) for v in description[1]):
                raise TypeError(
                    f"For string type {param_type}(aka {type_name}) with an enum, the"
                    " enum must be a list of strings."
                )
            return {
                "type": type_name,
                "description": description[0],
                "enum": description[1],
            }
        else:
            raise TypeError(
                "For string type, value must be a string containing the description of"
                " the field, or a tuple of length 2 where the first element is the"
                " description of the field and the second element is a list of strings"
                f" representing the enum. Got {description}."
            )
    else:
        if not isinstance(description, str):
            raise TypeError(
                f"For type {param_type}, value must be a"
                " string containing the description of the field."
            )
        return {"type": type_name, "description": description}


def get_tools_spec(func: Callable, name: Optional[str] = None) -> Dict[str, Any]:
    """
    Given a function stub, return a dictionary that is OpenAI tools api compatible
    for function calling. Note that since the OpenAI tools api is not explicitly
    documented, this is a best effort implementation.

    Args:
        func: a function, or a simple stub, which defines its parameters and properly
            annotates the types of the parameters, using Annotated[...], or default
            values to provide the description of the parameters.
        name: (optional) the name of the function. If not provided, the name of the
            function will be used.
    Returns:
        A dictionary that is OpenAI tools api compatible for function calling.
    """
    if not callable(func):
        raise TypeError("func must be a callable object.")
    function_name = name if name else func.__name__
    docstring = inspect.getdoc(func)
    # get the annotations of the function parameters
    type_hints = get_type_hints(func)
    # get the default values of the function parameters
    signature = inspect.signature(func)
    parameters = signature.parameters
    annotations = func.__annotations__

    # Constructing the JSON structure
    function_info = {
        "name": function_name,
        "description": docstring,
        "parameters": {"type": "object", "properties": {}},
    }

    # Adding parameter information to the JSON structure
    for param_name, param in parameters.items():
        # Determine the type of the parameter
        try:
            param_type = _original_type_backward_compatibility(type_hints[param_name])
        except KeyError:
            raise TypeError(f"Parameter {param_name} does not have a type annotation.")
        # Determine the annotation of the parameter
        param_annotation = annotations.get(param_name)
        # Determine the default value/description of the parameter
        default_value = param.default
        # Add parameter information to the JSON structure
        function_info["parameters"]["properties"][param_name] = _get_type_spec(
            param_type, param_annotation, default_value
        )

    return function_info



from ..conversable_agent import ConversableAgent
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from ....utils.client import ByzerLLM
from byzerllm.utils.retrieval import ByzerRetrieval
from ..agent import Agent
from ray.util.client.common import ClientActorHandle, ClientObjectRef
from .. import get_agent_name,run_agent_func,ChatResponse
from langchain import PromptTemplate

SYSTEM_PROMPT = '''You are a helpful AI assistant. The user will give you a conversation.
You should check the conversation is
'''
USER_PROMPT = '''User's question is: {input_question}

The conversation is: 

```
{input_context}
```
'''
class OutputAgent(ConversableAgent):
    

    def __init__(
        self,
        name: str,
        llm: ByzerLLM,        
        retrieval: ByzerRetrieval,                
        system_message: Optional[str] = SYSTEM_PROMPT,        
        is_termination_msg: Optional[Callable[[Dict], bool]] = None,
        max_consecutive_auto_reply: Optional[int] = None,
        human_input_mode: Optional[str] = "NEVER",
        code_execution_config: Optional[Union[Dict, bool]] = False,
        **kwargs,
    ):       
        super().__init__(
            name,
            llm,retrieval,
            system_message,
            is_termination_msg,
            max_consecutive_auto_reply,
            human_input_mode,
            code_execution_config=code_execution_config,            
            **kwargs,
        )

        self._reply_func_list = []
        # self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.generate_llm_reply)   
        self.register_reply([Agent, ClientActorHandle,str], OutputAgent.generate_converation_based_reply) 
        self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.check_termination_and_human_reply)

    def generate_retrieval_based_reply(
        self,
        raw_message: Optional[Union[Dict,str,ChatResponse]] = None,
        messages: Optional[List[Dict]] = None,
        sender: Optional[Union[ClientActorHandle,Agent,str]] = None,
        config: Optional[Any] = None,
    ) -> Tuple[bool, Union[str, Dict, None,ChatResponse]]:  
        
        box = []

        for message in messages:
            if not message["content"]:
                continue
            if message["role"] == "user":
                question = message["content"]
                box.append("Q:" + question)
            if message["role"] == "assistant":
                context = message["content"]   
                box.append("A:" + context)

        prompt = PromptTemplate.from_template(USER_PROMPT).format(input_question=box[0],input_context="\n".join(box))        
        new_message = {"content":prompt,"role":"user"}                    
        final,v = self.generate_llm_reply(None,[new_message],sender)
        return True,v + " TERMINATE"

                



import logging
import re
from typing import List, Tuple
import torch

import numpy as np
from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
    Pipeline,
    PreTrainedModel,
    PreTrainedTokenizer,
)

from transformers.utils import is_tf_available

if is_tf_available():
    import tensorflow as tf

from .consts import END_KEY, PROMPT_FOR_GENERATION_FORMAT, RESPONSE_KEY

logger = logging.getLogger(__name__)


def load_model_tokenizer_for_generate(
    pretrained_model_name_or_path: str,
    load_in_8bit: bool = False
) -> Tuple[PreTrainedModel, PreTrainedTokenizer]:
    """Loads the model and tokenizer so that it can be used for generating responses.

    Args:
        pretrained_model_name_or_path (str): name or path for model

    Returns:
        Tuple[PreTrainedModel, PreTrainedTokenizer]: model and tokenizer
    """
    tokenizer = AutoTokenizer.from_pretrained(pretrained_model_name_or_path, padding_side="left")
    model = AutoModelForCausalLM.from_pretrained(
        pretrained_model_name_or_path, 
        device_map="auto",
        load_in_8bit=load_in_8bit ,
        torch_dtype=torch.bfloat16, 
        trust_remote_code=True
    )
    return model, tokenizer


def get_special_token_id(tokenizer: PreTrainedTokenizer, key: str) -> int:
    """Gets the token ID for a given string that has been added to the tokenizer as a special token.

    When training, we configure the tokenizer so that the sequences like "### Instruction:" and "### End" are
    treated specially and converted to a single, new token.  This retrieves the token ID each of these keys map to.

    Args:
        tokenizer (PreTrainedTokenizer): the tokenizer
        key (str): the key to convert to a single token

    Raises:
        ValueError: if more than one ID was generated

    Returns:
        int: the token ID for the given key
    """
    token_ids = tokenizer.encode(key)
    if len(token_ids) > 1:
        raise ValueError(f"Expected only a single token for '{key}' but found {token_ids}")
    return token_ids[0]


class InstructionTextGenerationPipeline(Pipeline):
    def __init__(
        self, *args, do_sample: bool = True, max_new_tokens: int = 256, top_p: float = 0.92, top_k: int = 0, **kwargs
    ):
        """Initialize the pipeline

        Args:
            do_sample (bool, optional): Whether or not to use sampling. Defaults to True.
            max_new_tokens (int, optional): Max new tokens after the prompt to generate. Defaults to 128.
            top_p (float, optional): If set to float < 1, only the smallest set of most probable tokens with
                probabilities that add up to top_p or higher are kept for generation. Defaults to 0.92.
            top_k (int, optional): The number of highest probability vocabulary tokens to keep for top-k-filtering.
                Defaults to 0.
        """
        super().__init__(*args, do_sample=do_sample, max_new_tokens=max_new_tokens, top_p=top_p, top_k=top_k,
                         **kwargs)

    def _sanitize_parameters(self,
                             return_full_text: bool = None,
                             **generate_kwargs):
        preprocess_params = {}

        # newer versions of the tokenizer configure the response key as a special token.  newer versions still may
        # append a newline to yield a single token.  find whatever token is configured for the response key.
        tokenizer_response_key = next(
            (token for token in self.tokenizer.additional_special_tokens if token.startswith(RESPONSE_KEY)), None
        )

        response_key_token_id = None
        end_key_token_id = None
        if tokenizer_response_key:
            try:
                response_key_token_id = get_special_token_id(self.tokenizer, tokenizer_response_key)
                end_key_token_id = get_special_token_id(self.tokenizer, END_KEY)

                # Ensure generation stops once it generates "### End"
                generate_kwargs["eos_token_id"] = end_key_token_id
            except ValueError:
                pass

        forward_params = generate_kwargs
        postprocess_params = {
            "response_key_token_id": response_key_token_id,
            "end_key_token_id": end_key_token_id
        }

        if return_full_text is not None:
            postprocess_params["return_full_text"] = return_full_text

        return preprocess_params, forward_params, postprocess_params

    def preprocess(self, instruction_text, **generate_kwargs):
        prompt_text = PROMPT_FOR_GENERATION_FORMAT.format(instruction=instruction_text)
        inputs = self.tokenizer(
            prompt_text,
            return_tensors="pt",
        )
        inputs["prompt_text"] = prompt_text
        inputs["instruction_text"] = instruction_text
        return inputs

    def _forward(self, model_inputs, **generate_kwargs):
        input_ids = model_inputs["input_ids"]
        attention_mask = model_inputs.get("attention_mask", None)

        if input_ids.shape[1] == 0:
            input_ids = None
            attention_mask = None
            in_b = 1
        else:
            in_b = input_ids.shape[0]

        generated_sequence = self.model.generate(
            input_ids=input_ids.to(self.model.device),
            attention_mask=attention_mask,
            pad_token_id=self.tokenizer.pad_token_id,
            **generate_kwargs,
        )

        out_b = generated_sequence.shape[0]
        if self.framework == "pt":
            generated_sequence = generated_sequence.reshape(in_b, out_b // in_b, *generated_sequence.shape[1:])
        elif self.framework == "tf":
            generated_sequence = tf.reshape(generated_sequence, (in_b, out_b // in_b, *generated_sequence.shape[1:]))

        instruction_text = model_inputs.pop("instruction_text")
        return {"generated_sequence": generated_sequence, "input_ids": input_ids, "instruction_text": instruction_text}

    def postprocess(self, model_outputs, response_key_token_id, end_key_token_id, return_full_text: bool = False):

        generated_sequence = model_outputs["generated_sequence"][0]
        instruction_text = model_outputs["instruction_text"]

        generated_sequence: List[List[int]] = generated_sequence.numpy().tolist()
        records = []
        for sequence in generated_sequence:

            # The response will be set to this variable if we can identify it.
            decoded = None

            # If we have token IDs for the response and end, then we can find the tokens and only decode between them.
            if response_key_token_id and end_key_token_id:
                # Find where "### Response:" is first found in the generated tokens.  Considering this is part of the
                # prompt, we should definitely find it.  We will return the tokens found after this token.
                try:
                    response_pos = sequence.index(response_key_token_id)
                except ValueError:
                    logger.warn(f"Could not find response key {response_key_token_id} in: {sequence}")
                    response_pos = None

                if response_pos:
                    # Next find where "### End" is located.  The model has been trained to end its responses with this
                    # sequence (or actually, the token ID it maps to, since it is a special token).  We may not find
                    # this token, as the response could be truncated.  If we don't find it then just return everything
                    # to the end.  Note that even though we set eos_token_id, we still see the this token at the end.
                    try:
                        end_pos = sequence.index(end_key_token_id)
                    except ValueError:
                        end_pos = None

                    decoded = self.tokenizer.decode(sequence[response_pos + 1 : end_pos]).strip()

            if not decoded:
                # Otherwise we'll decode everything and use a regex to find the response and end.

                fully_decoded = self.tokenizer.decode(sequence)

                # The response appears after "### Response:".  The model has been trained to append "### End" at the
                # end.
                m = re.search(r"#+\s*Response:\s*(.+?)#+\s*End", fully_decoded, flags=re.DOTALL)

                if m:
                    decoded = m.group(1).strip()
                else:
                    # The model might not generate the "### End" sequence before reaching the max tokens.  In this case,
                    # return everything after "### Response:".
                    m = re.search(r"#+\s*Response:\s*(.+)", fully_decoded, flags=re.DOTALL)
                    if m:
                        decoded = m.group(1).strip()
                    else:
                        logger.warn(f"Failed to find response in:\n{fully_decoded}")

            # If the full text is requested, then append the decoded text to the original instruction.
            # This technically isn't the full text, as we format the instruction in the prompt the model has been
            # trained on, but to the client it will appear to be the full text.
            if return_full_text:
                decoded = f"{instruction_text}\n{decoded}"

            rec = {"generated_text": decoded}

            records.append(rec)

        return records


def generate_response(
    instruction: str,
    *,
    model: PreTrainedModel,
    tokenizer: PreTrainedTokenizer,
    **kwargs,
) -> str:
    """Given an instruction, uses the model and tokenizer to generate a response.  This formats the instruction in
    the instruction format that the model was fine-tuned on.

    Args:
        instruction (str): _description_
        model (PreTrainedModel): the model to use
        tokenizer (PreTrainedTokenizer): the tokenizer to use

    Returns:
        str: response
    """

    generation_pipeline = InstructionTextGenerationPipeline(model=model, tokenizer=tokenizer, **kwargs)
    return generation_pipeline(instruction)[0]["generated_text"]



import os
from os.path import expanduser
import urllib.request
import tarfile
from loguru import logger

import os
import re
from packaging import version
import json
from byzerllm.apps.byzer_storage.env import get_latest_byzer_retrieval_lib
from byzerllm.apps.byzer_storage import env

class StorageSubCommand:
        
    @staticmethod
    def install(args):
        version = args.version
        home = expanduser("~")        
        base_dir = args.base_dir or os.path.join(home, ".auto-coder")
        libs_dir = os.path.join(base_dir, "storage", "libs", f"byzer-retrieval-lib-{version}")
        if os.path.exists(libs_dir):
            print(f"Byzer Storage version {version} already installed.")
            return
        
        if not os.path.exists(base_dir):
            os.makedirs(base_dir,exist_ok=True)            

        env_info = env.detect_env()

        logger.info("Current environment:")
        logger.info(env_info)
        
        if env_info.java_version == "" or int(env_info.java_version) < 21:
            logger.info("JDK 21 not found, downloading and installing JDK 21...")
            try:
                env.download_and_install_jdk21(env_info, base_dir)
            except Exception as e:
                logger.error(f"Error downloading and installing JDK 21: {str(e)}. You may need to install JDK 21 manually.")
                        
        download_url = f"https://download.byzer.org/byzer-retrieval/byzer-retrieval-lib-{version}.tar.gz"
        libs_dir = os.path.join(base_dir, "storage", "libs")
        
        os.makedirs(libs_dir, exist_ok=True)
        download_path = os.path.join(libs_dir, f"byzer-retrieval-lib-{version}.tar.gz")
        if os.path.exists(download_path):
            logger.info(f"Byzer Storage version {version} already downloaded.")
        else:             
            def download_with_progressbar(url, filename):
                def progress(count, block_size, total_size):
                    percent = int(count * block_size * 100 / total_size)
                    print(f"\rDownload progress: {percent}%", end="")
            
                urllib.request.urlretrieve(url, filename,reporthook=progress)                
  
            logger.info(f"Download Byzer Storage version {version}: {download_url}")
            download_with_progressbar(download_url, download_path)    

            with tarfile.open(download_path, "r:gz") as tar:
                tar.extractall(path=libs_dir)
        
        print("Byzer Storage installed successfully")

    def collection(args):        
        from byzerllm.apps.llama_index.collection_manager import CollectionManager, CollectionItem
        home = expanduser("~")        
        base_dir = args.base_dir or os.path.join(home, ".auto-coder")
        collection_manager = CollectionManager(base_dir)
        if args.name:            
            collection = CollectionItem(name=args.name, description=args.description)
            collection_manager.add_collection(collection)
            print(f"Collection {args.name} added successfully.")
        else:
            print("Please provide collection name.")    
    
    @staticmethod
    def start(args):
        import byzerllm
        from byzerllm.utils.retrieval import ByzerRetrieval
        version = args.version
        cluster = args.cluster
        home = expanduser("~")        
        base_dir = args.base_dir or os.path.join(home, ".auto-coder")
        
        libs_dir = os.path.join(base_dir, "storage", "libs", f"byzer-retrieval-lib-{version}")
        data_dir = os.path.join(base_dir, "storage", "data")

        if not os.path.exists(os.path.join(data_dir,cluster)):
            os.makedirs(data_dir,exist_ok=True)
                
        if not os.path.exists(libs_dir):            
            StorageSubCommand.install(args)

        code_search_path = [libs_dir]
        
        logger.info(f"Connect and start Byzer Retrieval version {version}")
        env_vars = byzerllm.connect_cluster(address=args.ray_address,code_search_path=code_search_path)
        
        retrieval = ByzerRetrieval()
        retrieval.launch_gateway()

        if retrieval.is_cluster_exists(name=cluster):
            print(f"Cluster {cluster} exists already, stop it first.")
            return 
        
        cluster_json = os.path.join(base_dir, "storage", "data",f"{cluster}.json")
        if os.path.exists(cluster_json):
            StorageSubCommand.restore(args)
            print("Byzer Storage started successfully")
            return 
            
        builder = retrieval.cluster_builder()
        builder.set_name(cluster).set_location(data_dir).set_num_nodes(1).set_node_cpu(1).set_node_memory("2g")
        builder.set_java_home(env_vars["JAVA_HOME"]).set_path(env_vars["PATH"]).set_enable_zgc()
        builder.start_cluster()
        
        with open(os.path.join(base_dir, "storage", "data",f"{cluster}.json"),"w") as f:
            f.write(json.dumps(retrieval.cluster_info(cluster),ensure_ascii=False))

        print("Byzer Storage started successfully")

    @staticmethod 
    def stop(args):    
        import byzerllm
        from byzerllm.utils.retrieval import ByzerRetrieval
        version = args.version
        cluster = args.cluster
        home = expanduser("~")        
        base_dir = args.base_dir or os.path.join(home, ".auto-coder")
        libs_dir = os.path.join(base_dir, "storage", "libs", f"byzer-retrieval-lib-{version}")
        cluster_json = os.path.join(base_dir, "storage", "data",f"{cluster}.json")        

        if not os.path.exists(cluster_json) or not os.path.exists(libs_dir):
            print("No instance find.")
            return

        code_search_path = [libs_dir]
        
        logger.info(f"Connect and start Byzer Retrieval version {version}")
        byzerllm.connect_cluster(address=args.ray_address,code_search_path=code_search_path)             
        retrieval = ByzerRetrieval()
        retrieval.launch_gateway()
        retrieval.shutdown_cluster(cluster_name=cluster)

    @staticmethod 
    def export(args):   
        import byzerllm
        from byzerllm.utils.retrieval import ByzerRetrieval
        version = args.version
        cluster = args.cluster
        home = expanduser("~")        
        base_dir = args.base_dir or os.path.join(home, ".auto-coder")
        libs_dir = os.path.join(base_dir, "storage", "libs", f"byzer-retrieval-lib-{version}")
        cluster_json = os.path.join(base_dir, "storage", "data",f"{cluster}.json")        

        if not os.path.exists(cluster_json) or not os.path.exists(libs_dir):
            print("No instance find.")
            return

        code_search_path = [libs_dir]
        
        logger.info(f"Connect and restore Byzer Retrieval version {version}")
        byzerllm.connect_cluster(address=args.ray_address,code_search_path=code_search_path)        
     
        retrieval = ByzerRetrieval()
        retrieval.launch_gateway()
        
        with open(cluster_json,"w") as f:
            f.write(json.dumps(retrieval.cluster_info(cluster),ensure_ascii=False))

        print(f"Byzer Storage export successfully. Please check {cluster_json}")    


    
    def restore(args):
        import byzerllm
        from byzerllm.utils.retrieval import ByzerRetrieval
        version = args.version
        cluster = args.cluster
        home = expanduser("~")        
        base_dir = args.base_dir or os.path.join(home, ".auto-coder")
        libs_dir = os.path.join(base_dir, "storage", "libs", f"byzer-retrieval-lib-{version}")
        cluster_json = os.path.join(base_dir, "storage", "data",f"{cluster}.json")

        if not os.path.exists(cluster_json) or not os.path.exists(libs_dir):
            print("No instance find.")
            return

        code_search_path = [libs_dir]
        
        logger.info(f"Connect and restore Byzer Retrieval version {version}")
        byzerllm.connect_cluster(address=args.ray_address,code_search_path=code_search_path)        
     
        retrieval = ByzerRetrieval()
        retrieval.launch_gateway()

        if not retrieval.is_cluster_exists(cluster):
            with open(cluster_json,"r") as f:
                cluster_info = f.read()
            
            retrieval.restore_from_cluster_info(json.loads(cluster_info))
            
            print("Byzer Storage restore successfully")
        else:
            print(f"Cluster {cluster} is already exists")




import numpy as np
from typing import Union, List, Tuple, Optional, Dict


class Inference:
    def __init__(
        self,        
        model_dir: str,
        device_index:int = 0        
    ) -> None:
        from faster_whisper import WhisperModel
        self.model_dir = model_dir
        try:
            import ray
            import os
            print("ray.get_gpu_ids(): {}".format(ray.get_gpu_ids()))
            print("CUDA_VISIBLE_DEVICES: {}".format(os.environ["CUDA_VISIBLE_DEVICES"]))
        except Exception:
            pass    
        print(f"load whisper model: {self.model_dir}s cuda:{device_index}")
        self.model = WhisperModel(self.model_dir, device="cuda",compute_type="float16",device_index=device_index)

    def __call__(self, rate:int,t:np.ndarray, initial_prompt:str="以下是普通话的句子")->List[str]:        
        from scipy.io.wavfile import write as write_wav
        import io

        if t.dtype == "int64":
            t = t.astype(np.int32)

        byte_file = io.BytesIO()        
        write_wav(byte_file, rate, t)
        segments, info = self.model.transcribe(byte_file, beam_size=5,
        initial_prompt=initial_prompt
        )
        return [segment.text for segment in segments]


from typing import Dict, Optional, Union

import numpy as np

from .generation import GenerateModel
# codec_decode, generate_coarse, generate_fine, generate_text_semantic

class VoiceGenerateAPI:
    def __init__(self,model:GenerateModel) -> None:
        self.model = model

    def text_to_semantic(self,
        text: str,
        history_prompt: Optional[Union[Dict, str]] = None,
        temp: float = 0.7,
        silent: bool = False,
    ):
        """Generate semantic array from text.

        Args:
            text: text to be turned into audio
            history_prompt: history choice for audio cloning
            temp: generation temperature (1.0 more diverse, 0.0 more conservative)
            silent: disable progress bar

        Returns:
            numpy semantic array to be fed into `semantic_to_waveform`
        """
        x_semantic = self.model.generate_text_semantic(
            text,
            history_prompt=history_prompt,
            temp=temp,
            silent=silent,
            use_kv_caching=True
        )
        return x_semantic


    def semantic_to_waveform(self,
        semantic_tokens: np.ndarray,
        history_prompt: Optional[Union[Dict, str]] = None,
        temp: float = 0.7,
        silent: bool = False,
        output_full: bool = False,
    ):
        """Generate audio array from semantic input.

        Args:
            semantic_tokens: semantic token output from `text_to_semantic`
            history_prompt: history choice for audio cloning
            temp: generation temperature (1.0 more diverse, 0.0 more conservative)
            silent: disable progress bar
            output_full: return full generation to be used as a history prompt

        Returns:
            numpy audio array at sample frequency 24khz
        """
        coarse_tokens = self.model.generate_coarse(
            semantic_tokens,
            history_prompt=history_prompt,
            temp=temp,
            silent=silent,
            use_kv_caching=True
        )
        fine_tokens = self.model.generate_fine(
            coarse_tokens,
            history_prompt=history_prompt,
            temp=0.5,
        )
        audio_arr = self.model.codec_decode(fine_tokens)
        if output_full:
            full_generation = {
                "semantic_prompt": semantic_tokens,
                "coarse_prompt": coarse_tokens,
                "fine_prompt": fine_tokens,
            }
            return full_generation, audio_arr
        return audio_arr


    def save_as_prompt(self,filepath, full_generation):
        assert(filepath.endswith(".npz"))
        assert(isinstance(full_generation, dict))
        assert("semantic_prompt" in full_generation)
        assert("coarse_prompt" in full_generation)
        assert("fine_prompt" in full_generation)
        np.savez(filepath, **full_generation)


    def generate_audio(self,
        text: str,
        history_prompt: Optional[Union[Dict, str]] = None,
        text_temp: float = 0.7,
        waveform_temp: float = 0.7,
        silent: bool = False,
        output_full: bool = False,
    ):
        """Generate audio array from input text.

        Args:
            text: text to be turned into audio
            history_prompt: history choice for audio cloning
            text_temp: generation temperature (1.0 more diverse, 0.0 more conservative)
            waveform_temp: generation temperature (1.0 more diverse, 0.0 more conservative)
            silent: disable progress bar
            output_full: return full generation to be used as a history prompt

        Returns:
            numpy audio array at sample frequency 24khz
        """
        semantic_tokens = self.text_to_semantic(
            text,
            history_prompt=history_prompt,
            temp=text_temp,
            silent=silent,
        )
        out = self.semantic_to_waveform(
            semantic_tokens,
            history_prompt=history_prompt,
            temp=waveform_temp,
            silent=silent,
            output_full=output_full,
        )
        if output_full:
            full_generation, audio_arr = out
            return full_generation, audio_arr
        else:
            audio_arr = out
        return audio_arr



import argparse
import shlex
from byzerllm.lang import lang,locales

class StoreNestedDict(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        d = {}
        for kv in values:
            try:
                for clean_kv in shlex.split(kv):
                    key, value =clean_kv.split("=", 1)
                    d[key]=value
            except ValueError:
                raise argparse.ArgumentError(self, f"Invalid argument format: {kv}")
        setattr(namespace, self.dest, d)  

def get_command_args():
    parser = argparse.ArgumentParser(description=locales["desc"][lang])
    subparsers = parser.add_subparsers(dest='command')

    # Deploy 子命令
    deploy_cmd = subparsers.add_parser('deploy', help=locales["help_deploy"][lang])
    deploy_cmd.add_argument('--ray_address', default="auto", help=locales["help_ray_address"][lang])
    deploy_cmd.add_argument('--num_workers', type=int, default=1, help=locales["help_num_workers"][lang])
    deploy_cmd.add_argument('--worker_concurrency', type=int, default=1, help=locales["help_worker_concurrency"][lang])
    deploy_cmd.add_argument('--gpus_per_worker', type=float, default=0, help=locales["help_gpus_per_worker"][lang])
    deploy_cmd.add_argument('--cpus_per_worker', type=float, default=0.01, help=locales["help_cpus_per_worker"][lang])
    deploy_cmd.add_argument('--model_path', default="",required=False, help=locales["help_model_path"][lang])
    deploy_cmd.add_argument('--pretrained_model_type', default="custom/llama2", help=locales["help_pretrained_model_type"][lang])
    deploy_cmd.add_argument('--model', required=True, help=locales["help_udf_name"][lang])    
    deploy_cmd.add_argument('--infer_backend', default="", help=locales["help_infer_backend"][lang])
    deploy_cmd.add_argument('--infer_params', nargs='+', action=StoreNestedDict, help=locales["help_infer_params"][lang])
    deploy_cmd.add_argument("--file", default=None, required=False, help="")
    
    # Undeploy 子命令
    deploy_cmd = subparsers.add_parser('undeploy', help=locales["help_deploy"][lang])
    deploy_cmd.add_argument('--ray_address', default="auto", help=locales["help_ray_address"][lang])
    deploy_cmd.add_argument('--model', required=True, help=locales["help_udf_name"][lang])    
    deploy_cmd.add_argument("--file", default=None, required=False, help="")
    deploy_cmd.add_argument("--force",action='store_true', help="")

    # Query 子命令
    query_cmd = subparsers.add_parser('query', help=locales["help_query"][lang])
    query_cmd.add_argument('--ray_address', default="auto", help=locales["help_ray_address"][lang])
    query_cmd.add_argument('--model', required=True, help=locales["help_query_model"][lang])
    query_cmd.add_argument('--query', required=True, help=locales["help_query_text"][lang])
    query_cmd.add_argument('--template', default="auto", help=locales["help_template"][lang])
    query_cmd.add_argument("--file", default=None, required=False, help="")
    query_cmd.add_argument("--output_file", default="", help="")
    
    # Stat Command
    status_cmd = subparsers.add_parser('stat', help=locales["help_query"][lang])
    status_cmd.add_argument('--ray_address', default="auto", help=locales["help_ray_address"][lang])
    status_cmd.add_argument('--model', required=True, help=locales["help_query_model"][lang])        
    status_cmd.add_argument("--file", default=None, required=False, help="")    

    # Storage 子命令
    storage_cmd = subparsers.add_parser('storage', help='Manage Byzer Storage')    
    storage_cmd_subparsers = storage_cmd.add_subparsers(dest='storage_command')
    
    # install 子命令
    storage_install_cmd = storage_cmd_subparsers.add_parser('install', help='Install Byzer Storage')
    storage_install_cmd.add_argument("--file", default=None, required=False, help="")
    storage_install_cmd.add_argument('--ray_address', default="auto", help=locales["help_ray_address"][lang])
    storage_install_cmd.add_argument("--version", default="0.1.11", required=False, help="")
    storage_install_cmd.add_argument('--cluster', default="byzerai_store", help="")
    storage_install_cmd.add_argument('--base_dir', default="", help="")

    storage_collection_cmd = storage_cmd_subparsers.add_parser('collection', help='')
    storage_collection_cmd.add_argument("--file", default=None, required=False, help="")
    storage_collection_cmd.add_argument('--ray_address', default="auto", help=locales["help_ray_address"][lang])
    storage_collection_cmd.add_argument("--version", default="0.1.11", required=False, help="")
    storage_collection_cmd.add_argument('--cluster', default="byzerai_store", help="")
    storage_collection_cmd.add_argument('--base_dir', default="", help="")
    storage_collection_cmd.add_argument('--name', default="", help="")
    storage_collection_cmd.add_argument('--description', default="", help="")
    
    # start 子命令
    storage_start_command = storage_cmd_subparsers.add_parser('start', help='Start Byzer Storage')
    storage_start_command.add_argument("--file", default=None, required=False, help="")
    storage_start_command.add_argument("--version", default="0.1.11", required=False, help="")
    storage_start_command.add_argument('--ray_address', default="auto", help=locales["help_ray_address"][lang])
    storage_start_command.add_argument('--cluster', default="byzerai_store", help="")
    storage_start_command.add_argument('--base_dir', default="", help="")


    # stop 子命令
    storage_stop_command = storage_cmd_subparsers.add_parser('stop', help='Stop Byzer Storage')
    storage_stop_command.add_argument("--file", default=None, required=False, help="")
    storage_stop_command.add_argument("--version", default="0.1.11", required=False, help="")
    storage_stop_command.add_argument('--ray_address', default="auto", help=locales["help_ray_address"][lang])
    storage_stop_command.add_argument('--cluster', default="byzerai_store", help="")
    storage_stop_command.add_argument('--base_dir', default="", help="")
    
    # export 子命令
    storage_export_command = storage_cmd_subparsers.add_parser('export', help='Export Byzer Storage Information')    
    storage_export_command.add_argument("--file", default=None, required=False, help="")
    storage_export_command.add_argument("--version", default="0.1.11", required=False, help="")
    storage_export_command.add_argument('--ray_address', default="auto", help=locales["help_ray_address"][lang])
    storage_export_command.add_argument('--cluster', default="byzerai_store", help="")
    storage_export_command.add_argument('--base_dir', default="", help="")
    
    # Serve 子命令
    serve_cmd = subparsers.add_parser('serve', help='Serve deployed models with OpenAI compatible APIs')    
    serve_cmd.add_argument("--file", default="", help="")        
    serve_cmd.add_argument("--ray_address", default="auto", help="")    
    serve_cmd.add_argument("--host", default="", help="")
    serve_cmd.add_argument("--port", type=int, default=8000, help="")
    serve_cmd.add_argument("--uvicorn_log_level", default="info", help="")
    serve_cmd.add_argument("--allow_credentials", action='store_true', help="")
    serve_cmd.add_argument("--allowed_origins", default=["*"], help="")
    serve_cmd.add_argument("--allowed_methods", default=["*"], help="")
    serve_cmd.add_argument("--allowed_headers", default=["*"], help="")
    serve_cmd.add_argument("--api_key", default="", help="")
    serve_cmd.add_argument("--served_model_name", default="", help="")
    serve_cmd.add_argument("--prompt_template", default="", help="")
    serve_cmd.add_argument("--ssl_keyfile", default="", help="")
    serve_cmd.add_argument("--ssl_certfile", default="", help="")
    serve_cmd.add_argument("--response_role", default="assistant", help="")
    serve_cmd.add_argument('--template', default="auto", help=locales["help_template"][lang])

    return parser.parse_args()



from langchain.text_splitter import RecursiveCharacterTextSplitter
from typing import Any, List, Dict
import os
import json
import shutil
from langchain.vectorstores import FAISS
from langchain.docstore.document import Document
from pathlib import Path
from . import BuilderParams


class OnceWay:
    def __init__(self,db_dir,embeddings) -> None:
        self.db_dir = db_dir
        self.embeddings = embeddings

    def build(self,path:str,params:BuilderParams,extra_params={}):
        p = Path(path)
        docs = []
        items = list(p.rglob("**/*.json"))                             

        for i in items:
            if i.is_file():
               with open(i.as_posix(),"r",encoding="utf-8") as f:
                 for line in f:
                    doc = json.loads(line)
                    docs.append(Document(page_content=doc["page_content"],
                                         metadata={"source": doc["source"], "page_content": doc["page_content"]}))
                    
        
        if  len(docs) > 0:
            text_splitter = RecursiveCharacterTextSplitter(chunk_size=params.chunk_size, chunk_overlap=params.chunk_overlap)
            split_docs = text_splitter.split_documents(docs)             
            print(f"Build vector db in {self.db_dir}. total docs: {len(docs)} total split docs: {len(split_docs)}")
            db = FAISS.from_documents(split_docs, self.embeddings)                                    
            db.save_local(self.db_dir)
            docs.clear()               


class MergeWay:
    def __init__(self,db_dir,embeddings) -> None:
        self.db_dir = db_dir
        self.embeddings = embeddings

    def build(self,path:str,params:BuilderParams,extra_params={}):
        p = Path(path)
        docs = []
        items = list(p.rglob("**/*.json"))

        max_doc_size = params.batch_size

        paths = []
        counter = 0        

        for i in items:
            if i.is_file():
               with open(i.as_posix(),"r",encoding="utf-8") as f:
                 for line in f:
                    doc = json.loads(line)                    
                    docs.append(Document(page_content=doc["page_content"],metadata={ "source":doc["source"]}))
                    if len(docs) > max_doc_size:
                        text_splitter = RecursiveCharacterTextSplitter(chunk_size=params.chunk_size, chunk_overlap=params.chunk_overlap)
                        split_docs = text_splitter.split_documents(docs) 
                        print(f"Build vector db in {self.db_dir}. total docs: {len(docs)} total split docs: {len(split_docs)}")
                        db = FAISS.from_documents(split_docs, self.embeddings)
                        counter += 1
                        temp_path = os.path.join(self.db_dir,str(counter))
                        paths.append(temp_path)
                        db.save_local(temp_path)
                        docs.clear()

        if  len(docs) > 0:
            text_splitter = RecursiveCharacterTextSplitter(chunk_size=params.chunk_size, chunk_overlap=params.chunk_overlap)
            split_docs = text_splitter.split_documents(docs) 
            print(f"Build vector db in {self.db_dir}. total docs: {len(docs)} total split docs: {len(split_docs)}")
            db = FAISS.from_documents(split_docs, self.embeddings)
            counter += 1
            temp_path = os.path.join(self.db_dir,str(counter))
            paths.append(temp_path)
            db.save_local(temp_path)
            docs.clear()
        
        print(f"load vector index from {paths[0]}")
        t_db = FAISS.load_local(paths[0],self.embeddings)
        for item in paths[1:]:
            print(f"merge vector index from {paths[0]}")
            t_db.merge_from(FAISS.load_local(item,self.embeddings))
        t_db.save_local(self.db_dir)
        
        print(F"clean temp file...")
        for p in paths:
            shutil.rmtree(p)



import ray
import os

from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from byzerllm.utils import generate_str_md5
from byzerllm.utils.client import ByzerLLM,default_chat_wrapper,LLMResponse
from byzerllm.utils.retrieval import ByzerRetrieval
from ray.util.client.common import ClientActorHandle, ClientObjectRef

from byzerllm.apps.agent import Agent,Agents,get_agent_name,run_agent_func,ChatResponse,modify_message_metadata,modify_message_content

from byzerllm.apps.agent.user_proxy_agent import UserProxyAgent
from byzerllm.apps.agent.extensions.data_analysis_pipeline_agent import DataAnalysisPipeline,DataAnalysisPipelineManager
from byzerllm.apps.agent.extensions.simple_retrieval_client import SimpleRetrievalClient
from byzerllm.apps.agent.store.memory_store import  MessageStore,MemoryStore,Message as ChatStoreMessage
from byzerllm.apps.agent.store.stores import Stores
try:
    from termcolor import colored
except ImportError:
    def colored(x, *args, **kwargs):
        return x


class DataAnalysis:
    def __init__(self,chat_name:str, 
                 owner:str,
                 file_path:str,
                 llm:ByzerLLM,
                 retrieval:ByzerRetrieval,
                 use_shared_disk:bool=False, 
                 chat_wrapper = default_chat_wrapper ,
                 skip_preview_file:bool=False,
                 retrieval_cluster:str="data_analysis",
                 retrieval_db:str="data_analysis",
                 message_store:Optional[Optional[Union[str,ClientActorHandle,MessageStore]]]=None,              
                 ):
        self.chat_name = chat_name
        self.owner = owner
        self.chat_wrapper = chat_wrapper
        self.suffix = generate_str_md5(f"{self.chat_name}_{self.owner}")
        self.name = f"data_analysis_pp_{self.suffix}"   
        self.manager = self.get_pipeline_manager() 

        self.message_store = message_store 
        
        self.use_shared_disk = use_shared_disk
        self.llm = llm
        self.retrieval = retrieval
        
        self.retrieval_cluster = retrieval_cluster
        self.retrieval_db = retrieval_db

        self.simple_retrieval_client = SimpleRetrievalClient(llm=self.llm,
                                                        retrieval=self.retrieval,
                                                        retrieval_cluster=self.retrieval_cluster,
                                                        retrieval_db=self.retrieval_db,
                                                        )     
        self.file_path = file_path
        self.file_ref = None   

        if self.message_store:
            if isinstance(self.message_store,str):
                try:
                    ray.get_actor(self.message_store)
                except:
                    ray.remote(MemoryStore).options(num_cpus=0.1,name=self.message_store, lifetime="detached").remote()


        if not ray.get(self.manager.check_pipeline_exists.remote(self.name)):
            if self.file_path and not self.use_shared_disk:
                base_name = os.path.basename(file_path)
                _, ext = os.path.splitext(base_name)
                new_base_name = self.name + ext
                dir_name = os.path.dirname(file_path)
                new_file_path = os.path.join(dir_name, new_base_name)
                print(f"use_shared_disk: {self.use_shared_disk} file_path: {self.file_path} new_file_path: {new_file_path}",flush=True)
                self.file_ref = ray.put(open(self.file_path,"rb").read())
                self.file_path = new_file_path

            self.data_analysis_pipeline = ray.get(self.manager.get_or_create_pipeline.remote(
                name = self.name,
                llm =llm,
                retrieval =retrieval,
                file_path=self.file_path,
                file_ref=self.file_ref,
                chat_name = self.chat_name,
                owner = self.owner,
                chat_wrapper = self.chat_wrapper,
                message_store = self.message_store,
                )) 

            # trigger file preview manually
            if not skip_preview_file:
                ray.get(self.data_analysis_pipeline.preview_file.remote()) 
        else:
            self.data_analysis_pipeline = ray.get(self.manager.get_pipeline.remote(self.name))

        self.client = self.get_or_create_user(f"user_{self.name}")
        
                

    def get_or_create_user(self,name:str)->bool:
        try:
            return ray.get_actor(name)            
        except Exception:
            return Agents.create_remote_agent(UserProxyAgent,f"user_{self.name}",self.llm,self.retrieval,
                                human_input_mode="NEVER",
                                max_consecutive_auto_reply=0,chat_wrapper=self.chat_wrapper)
        
    def get_messages(self):
        v = [] 
        store = Stores("MESSAGE_STORE")
        for item in store.get(self.name):            
            v.append(item.sender + " (to " + f"{item.receiver}):\n")
            v.append(f"{item.m['content']}")
            v.append("\n" + "-" * 80)
        return "\n".join(v)    

    def send_from_agent_to_agent(self,from_agent_name:str,to_agent_name:str,message:Dict[str,Any]):
        if self.data_analysis_pipeline is None:
            return None
        return ray.get(self.data_analysis_pipeline.send_from_agent_to_agent.remote(from_agent_name,to_agent_name,message))  
        
    def analyze(self,content:str,metadata:Dict[str,Any]={}): 
        if self.message_store is not None:
            store = Stores("MESSAGE_STORE")
            store.clear(self.name)

        ray.get(self.data_analysis_pipeline.update_max_consecutive_auto_reply.remote(1))
        ray.get(           
           self.client.initiate_chat.remote(
                self.data_analysis_pipeline,
                message={
                    "content":content,
                    "role":"user",
                    "metadata":{  
                        **metadata                      
                    }                    
                },
           ) 
        ) 
        output = self.output()   
        
        # if self.message_store is not None:
        #     self.message_store.put(ChatStoreMessage(self.name,None,None,None,-1))
        metadata = output["metadata"]        
        
        if "stream" in metadata and metadata["stream"]:
            agent = metadata["agent"]
            stream_id = metadata["stream_id"]
            result = []
            yield metadata["contexts"]
            for item in self.data_analysis_pipeline.get_agent_stream_messages.remote(agent,stream_id):
                t = ray.get(item)
                result.append(t[0])
                yield t
            
            if self.retrieval:
                self.simple_retrieval_client.save_conversation(owner=self.owner,chat_name=self.chat_name,role="user",content=content)
                r = "".join(result)
                self.simple_retrieval_client.save_conversation(owner=self.owner,chat_name=self.chat_name,role="assistant",content=r) 
      
        else:    
            if self.retrieval:
                self.simple_retrieval_client.save_conversation(owner=self.owner,chat_name=self.chat_name,role="user",content=content)
                self.simple_retrieval_client.save_conversation(owner=self.owner,chat_name=self.chat_name,role="assistant",content=output["content"])
            return output

    def get_chat_messages(self):
        return ray.get(self.data_analysis_pipeline.get_chat_messages.remote())   

    def close(self):        
        try:
            ray.kill(ray.get_actor(f"user_{self.name}"))
        except Exception:
            pass
        
        try:            
            ray.kill(ray.get_actor(f"{self.name}"))
        except Exception:
            pass

        try:            
            ray.get(self.manager.remove_pipeline.remote(self.name))  
        except Exception:
            pass

        self.data_analysis_pipeline = None                      
    
    def output(self):
        return ray.get(self.data_analysis_pipeline.last_message.remote(get_agent_name(self.client)))  

    def update_pipeline_system_message(self,system_message:str)->bool: 
        if self.data_analysis_pipeline is None:
            return False           
        ray.get(self.data_analysis_pipeline.update_system_message.remote(system_message))
        return True

    def update_agent_system_message(self,agent_name:str,system_message:str)->bool:
        if self.data_analysis_pipeline is None:
            return False 
        return ray.get(self.data_analysis_pipeline.update_system_message_by_agent.remote(agent_name,system_message))        
    
    def get_agent_system_message(self,agent_name:str)->str:
        if self.data_analysis_pipeline is None:
            return ""
        return ray.get(self.data_analysis_pipeline.get_agent_system_message.remote(agent_name))
    
    def get_pipeline_system_message(self)->str:
        if self.data_analysis_pipeline is None:
            return ""
        return ray.get(self.data_analysis_pipeline.get_system_message.remote())
    
    def get_agent_names(self):
        if self.data_analysis_pipeline is None:
            return []
        return ray.get(self.data_analysis_pipeline.get_agent_names.remote())
    
    def clear_agent_message_box(self,agent_name:str,last_n=0)->bool:
        if self.data_analysis_pipeline is None:
            return False
        return ray.get(self.data_analysis_pipeline.clear_agent_message_box.remote(agent_name,last_n))

    
    def get_pipeline_manager(self)->ClientActorHandle:
        name = "DATA_ANALYSIS_PIPELINE_MANAGER"
        manager = None
        try:
            manager = ray.get_actor(name)
        except Exception:              
            manager = ray.remote(DataAnalysisPipelineManager).options(
                name=name, 
                lifetime="detached", 
                max_concurrency=500,              
                num_cpus=1,
                num_gpus=0
            ).remote()
        return manager     
        
        
    
        


import logging
import os
import pathlib
import re
import subprocess
import sys
import time
import json
from concurrent.futures import ThreadPoolExecutor, TimeoutError
from hashlib import md5
from typing import Callable, Dict, List, Optional, Tuple, Union
try:
    import docker
except ImportError:
    docker = None



# Regular expression for finding a code block
DEFAULT_MODEL="chat"
CODE_BLOCK_PATTERN = r"```[ \t]*(\w+)?[ \t]*\r?\n(.*?)\r?\n[ \t]*```"
WORKING_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), "extensions")
UNKNOWN = "unknown"
TIMEOUT_MSG = "Timeout"
DEFAULT_TIMEOUT = 600
WIN32 = sys.platform == "win32"
PATH_SEPARATOR = WIN32 and "\\" or "/"

logger = logging.getLogger(__name__)


def get_value_from_llm_str(v:str,k:str, default_value)->Union[str,int,float,bool,None]:
    responses = extract_code(v)
    value = default_value
    for lang,code in responses:
        if lang == "json":
            try:
                value = json.loads(code)[k]
            except Exception as inst:
                pass 
    return value

def content_str(content: Union[str, List, None]) -> str:
    """Converts `content` into a string format.

    This function processes content that may be a string, a list of mixed text and image URLs, or None,
    and converts it into a string. Text is directly appended to the result string, while image URLs are
    represented by a placeholder image token. If the content is None, an empty string is returned.

    Args:
        - content (Union[str, List, None]): The content to be processed. Can be a string, a list of dictionaries
                                      representing text and image URLs, or None.

    Returns:
        str: A string representation of the input content. Image URLs are replaced with an image token.

    Note:
    - The function expects each dictionary in the list to have a "type" key that is either "text" or "image_url".
      For "text" type, the "text" key's value is appended to the result. For "image_url", an image token is appended.
    - This function is useful for handling content that may include both text and image references, especially
      in contexts where images need to be represented as placeholders.
    """
    if content is None:
        return ""
    if isinstance(content, str):
        return content
    if not isinstance(content, list):
        raise TypeError(f"content must be None, str, or list, but got {type(content)}")

    rst = ""
    for item in content:
        if not isinstance(item, dict):
            raise TypeError("Wrong content format: every element should be dict if the content is a list.")
        assert "type" in item, "Wrong content format. Missing 'type' key in content's dict."
        if item["type"] == "text":
            rst += item["text"]
        elif item["type"] == "image_url":
            rst += "<image>"
        else:
            raise ValueError(f"Wrong content format: unknown type {item['type']} within the content")
    return rst

def infer_lang(code):
    """infer the language for the code.
    TODO: make it robust.
    """
    if code.startswith("python ") or code.startswith("pip") or code.startswith("python3 "):
        return "sh"

    # check if code is a valid python code
    try:
        compile(code, "test", "exec")
        return "python"
    except SyntaxError:
        # not a valid python code
        return UNKNOWN


def check_target_codes_exists(codes: List[Tuple[str, str]], langs: List[str]) -> bool:
    """Check if there is code in a specific language in the code list.

    Args:
        codes (list): The list of code blocks.
        langs (list): The language to check.

    Returns:
        bool: True if there is code in the specified language; False otherwise.
    """
    for l, _ in codes:
        if l in langs:
            return True
    return False

def get_target_codes(codes: List[Tuple[str, str]], langs: List[str]) -> List[str]:
    """Get code in a specific language from the code list.

    Args:
        codes (list): The list of code blocks.
        langs (list): The language to check.

    Returns:
        str: The code in the specified language.
    """
    target_codes = []
    for l, code in codes:
        if l in langs:
            target_codes.append(code)
    return target_codes


def extract_code(
    text: Union[str, List], pattern: str = CODE_BLOCK_PATTERN, detect_single_line_code: bool = False
) -> List[Tuple[str, str]]:
    """Extract code from a text.

    Args:
        text (str or List): The content to extract code from. The content can be
            a string or a list, as returned by standard GPT or multimodal GPT.
        pattern (str, optional): The regular expression pattern for finding the
            code block. Defaults to CODE_BLOCK_PATTERN.
        detect_single_line_code (bool, optional): Enable the new feature for
            extracting single line code. Defaults to False.

    Returns:
        list: A list of tuples, each containing the language and the code.
          If there is no code block in the input text, the language would be "unknown".
          If there is code block but the language is not specified, the language would be "".
    """
    text = content_str(text)
    if not detect_single_line_code:
        match = re.findall(pattern, text, flags=re.DOTALL)
        return match if match else [(UNKNOWN, text)]

    # Extract both multi-line and single-line code block, separated by the | operator
    # `([^`]+)`: Matches inline code.
    code_pattern = re.compile(CODE_BLOCK_PATTERN + r"|`([^`]+)`")
    code_blocks = code_pattern.findall(text)

    # Extract the individual code blocks and languages from the matched groups
    extracted = []
    for lang, group1, group2 in code_blocks:
        if group1:
            extracted.append((lang.strip(), group1.strip()))
        elif group2:
            extracted.append(("", group2.strip()))

    return extracted


def _cmd(lang):
    if lang.startswith("python") or lang in ["bash", "sh", "powershell"]:
        return lang
    if lang in ["shell"]:
        return "sh"
    if lang in ["ps1"]:
        return "powershell"
    raise NotImplementedError(f"{lang} not recognized in code execution")


def execute_code(
    code: Optional[str] = None,
    timeout: Optional[int] = None,
    filename: Optional[str] = None,
    work_dir: Optional[str] = None,
    use_docker: Optional[Union[List[str], str, bool]] = None,
    lang: Optional[str] = "python",
) -> Tuple[int, str, str]:
    """Execute code in a docker container.
    This function is not tested on MacOS.

    Args:
        code (Optional, str): The code to execute.
            If None, the code from the file specified by filename will be executed.
            Either code or filename must be provided.
        timeout (Optional, int): The maximum execution time in seconds.
            If None, a default timeout will be used. The default timeout is 600 seconds. On Windows, the timeout is not enforced when use_docker=False.
        filename (Optional, str): The file name to save the code or where the code is stored when `code` is None.
            If None, a file with a randomly generated name will be created.
            The randomly generated file will be deleted after execution.
            The file name must be a relative path. Relative paths are relative to the working directory.
        work_dir (Optional, str): The working directory for the code execution.
            If None, a default working directory will be used.
            The default working directory is the "extensions" directory under
            "path_to_autogen".
        use_docker (Optional, list, str or bool): The docker image to use for code execution.
            If a list or a str of image name(s) is provided, the code will be executed in a docker container
            with the first image successfully pulled.
            If None, False or empty, the code will be executed in the current environment.
            Default is None, which will be converted into an empty list when docker package is available.
            Expected behaviour:
                - If `use_docker` is explicitly set to True and the docker package is available, the code will run in a Docker container.
                - If `use_docker` is explicitly set to True but the Docker package is missing, an error will be raised.
                - If `use_docker` is not set (i.e., left default to None) and the Docker package is not available, a warning will be displayed, but the code will run natively.
            If the code is executed in the current environment,
            the code must be trusted.
        lang (Optional, str): The language of the code. Default is "python".

    Returns:
        int: 0 if the code executes successfully.
        str: The error message if the code fails to execute; the stdout otherwise.
        image: The docker image name after container run when docker is used.
    """    
    if all((code is None, filename is None)):
        error_msg = f"Either {code=} or {filename=} must be provided."
        logger.error(error_msg)
        raise AssertionError(error_msg)

    # Warn if use_docker was unspecified (or None), and cannot be provided (the default).
    # In this case the current behavior is to fall back to run natively, but this behavior
    # is subject to change.
    if use_docker is None:
        if docker is None:
            use_docker = False
            logger.warning(
                "execute_code was called without specifying a value for use_docker. Since the python docker package is not available, code will be run natively. Note: this fallback behavior is subject to change"
            )
        else:
            # Default to true
            use_docker = True

    timeout = timeout or DEFAULT_TIMEOUT
    original_filename = filename
    if WIN32 and lang in ["sh", "shell"] and (not use_docker):
        lang = "ps1"
    if filename is None:
        code_hash = md5(code.encode()).hexdigest()
        # create a file with a automatically generated name
        filename = f"tmp_code_{code_hash}.{'py' if lang.startswith('python') else lang}"
    if work_dir is None:
        work_dir = WORKING_DIR
    filepath = os.path.join(work_dir, filename)
    file_dir = os.path.dirname(filepath)
    os.makedirs(file_dir, exist_ok=True)
    if code is not None:
        with open(filepath, "w", encoding="utf-8") as fout:
            fout.write(code)
    # check if already running in a docker container
    in_docker_container = os.path.exists("/.dockerenv")
    if not use_docker or in_docker_container:
        # already running in a docker container
        cmd = [
            sys.executable if lang.startswith("python") else _cmd(lang),
            f".\\{filename}" if WIN32 else filename,
        ]
        if WIN32:
            logger.warning("SIGALRM is not supported on Windows. No timeout will be enforced.")
            result = subprocess.run(
                cmd,
                cwd=work_dir,
                capture_output=True,
                text=True,
            )
        else:
            with ThreadPoolExecutor(max_workers=1) as executor:
                future = executor.submit(
                    subprocess.run,
                    cmd,
                    cwd=work_dir,
                    capture_output=True,
                    text=True,
                )
                try:
                    result = future.result(timeout=timeout)
                except TimeoutError:
                    if original_filename is None:
                        os.remove(filepath)
                    return 1, TIMEOUT_MSG, None
        if original_filename is None:
            os.remove(filepath)
        if result.returncode:
            logs = result.stderr
            if original_filename is None:
                abs_path = str(pathlib.Path(filepath).absolute())
                logs = logs.replace(str(abs_path), "").replace(filename, "")
            else:
                abs_path = str(pathlib.Path(work_dir).absolute()) + PATH_SEPARATOR
                logs = logs.replace(str(abs_path), "")
        else:
            logs = result.stdout
        return result.returncode, logs, None

    # create a docker client
    client = docker.from_env()
    image_list = (
        ["python:3-alpine", "python:3", "python:3-windowsservercore"]
        if use_docker is True
        else [use_docker]
        if isinstance(use_docker, str)
        else use_docker
    )
    for image in image_list:
        # check if the image exists
        try:
            client.images.get(image)
            break
        except docker.errors.ImageNotFound:
            # pull the image
            print("Pulling image", image)
            try:
                client.images.pull(image)
                break
            except docker.errors.DockerException:
                print("Failed to pull image", image)
    # get a randomized str based on current time to wrap the exit code
    exit_code_str = f"exitcode{time.time()}"
    abs_path = pathlib.Path(work_dir).absolute()
    cmd = [
        "sh",
        "-c",
        f"{_cmd(lang)} {filename}; exit_code=$?; echo -n {exit_code_str}; echo -n $exit_code; echo {exit_code_str}",
    ]
    # create a docker container
    container = client.containers.run(
        image,
        command=cmd,
        working_dir="/workspace",
        detach=True,
        # get absolute path to the working directory
        volumes={abs_path: {"bind": "/workspace", "mode": "rw"}},
    )
    start_time = time.time()
    while container.status != "exited" and time.time() - start_time < timeout:
        # Reload the container object
        container.reload()
    if container.status != "exited":
        container.stop()
        container.remove()
        if original_filename is None:
            os.remove(filepath)
        return 1, TIMEOUT_MSG, image
    # get the container logs
    logs = container.logs().decode("utf-8").rstrip()
    # commit the image
    tag = filename.replace("/", "")
    container.commit(repository="python", tag=tag)
    # remove the container
    container.remove()
    # check if the code executed successfully
    exit_code = container.attrs["State"]["ExitCode"]
    if exit_code == 0:
        # extract the exit code from the logs
        pattern = re.compile(f"{exit_code_str}(\\d+){exit_code_str}")
        match = pattern.search(logs)
        exit_code = 1 if match is None else int(match.group(1))
        # remove the exit code from the logs
        logs = logs if match is None else pattern.sub("", logs)

    if original_filename is None:
        os.remove(filepath)
    if exit_code:
        logs = logs.replace(f"/workspace/{filename if original_filename is None else ''}", "")
    # return the exit code, logs and image
    return exit_code, logs, f"python:{tag}"


def _remove_check(response):
    """Remove the check function from the response."""
    # find the position of the check function
    pos = response.find("def check(")
    if pos == -1:
        return response
    return response[:pos]


def eval_function_completions(
    responses: List[str],
    definition: str,
    test: Optional[str] = None,
    entry_point: Optional[str] = None,
    assertions: Optional[Union[str, Callable[[str], Tuple[str, float]]]] = None,
    timeout: Optional[float] = 3,
    use_docker: Optional[bool] = True,
) -> Dict:
    """(openai<1) Select a response from a list of responses for the function completion task (using generated assertions), and/or evaluate if the task is successful using a gold test.

    Args:
        responses (list): The list of responses.
        definition (str): The input definition.
        test (Optional, str): The test code.
        entry_point (Optional, str): The name of the function.
        assertions (Optional, str or Callable): The assertion code which serves as a filter of the responses, or an assertion generator.
            When provided, only the responses that pass the assertions will be considered for the actual test (if provided).
        timeout (Optional, float): The timeout for executing the code.

    Returns:
        dict: The success metrics.
    """
    n = len(responses)
    if assertions is None:
        # no assertion filter
        success_list = []
        for i in range(n):
            response = _remove_check(responses[i])
            code = (
                f"{response}\n{test}\ncheck({entry_point})"
                if response.startswith("def")
                else f"{definition}{response}\n{test}\ncheck({entry_point})"
            )
            success = execute_code(code, timeout=timeout, use_docker=use_docker)[0] == 0
            success_list.append(success)
        return {
            "expected_success": 1 - pow(1 - sum(success_list) / n, n),
            "success": any(s for s in success_list),
        }
    if callable(assertions) and n > 1:
        # assertion generator
        assertions, gen_cost = assertions(definition)
    else:
        assertions, gen_cost = None, 0
    if n > 1 or test is None:
        for i in range(n):
            response = responses[i] = _remove_check(responses[i])
            code = (
                f"{response}\n{assertions}" if response.startswith("def") else f"{definition}{response}\n{assertions}"
            )
            succeed_assertions = execute_code(code, timeout=timeout, use_docker=use_docker)[0] == 0
            if succeed_assertions:
                break
    else:
        # just test, no need to check assertions
        succeed_assertions = False
        i, response = 0, responses[0]
    if test is None:
        # no test code
        return {
            "index_selected": i,
            "succeed_assertions": succeed_assertions,
            "gen_cost": gen_cost,
            "assertions": assertions,
        }
    code_test = (
        f"{response}\n{test}\ncheck({entry_point})"
        if response.startswith("def")
        else f"{definition}{response}\n{test}\ncheck({entry_point})"
    )
    success = execute_code(code_test, timeout=timeout, use_docker=use_docker)[0] == 0
    return {
        "index_selected": i,
        "succeed_assertions": succeed_assertions,
        "success": success,
        "gen_cost": gen_cost,
        "assertions": assertions,
    }


_FUNC_COMPLETION_PROMPT = "# Python 3{definition}"
_FUNC_COMPLETION_STOP = ["\nclass", "\ndef", "\nif", "\nprint"]

class PassAssertionFilter:
    def __init__(self, assertions):
        self._assertions = assertions        
        self.cost = 0
        self.metrics = self.responses = None

    def pass_assertions(self, context, response, **_):        
        responses = response[0].output        
        metrics = eval_function_completions(responses, context["definition"], assertions=self._assertions)
        self._assertions = metrics["assertions"]
        self.cost += metrics["gen_cost"]
        self.metrics = metrics
        self.responses = responses
        return metrics["succeed_assertions"]




import json
import time
from abc import ABC
from typing import List, Tuple, Dict, Optional

import boto3

from byzerllm.log import init_logger
from byzerllm.utils import random_uuid
from byzerllm.utils.langutil import asyncfy_with_semaphore

logger = init_logger(__name__)


class SupportModelProviders:
    Anthropic = "anthropic"
    Meta = "meta"


class CustomSaasAPI:
    def __init__(self, infer_params: Dict[str, str]) -> None:
        self.aws_access_key = infer_params["saas.aws_access_key"]
        self.aws_secret_key = infer_params["saas.aws_secret_key"]
        self.region_name = infer_params["saas.region_name"]
        self.model = infer_params["saas.model"]
        self.model_api_version = infer_params.get("saas.model_api_version", None)
        self.model_provider = self.model.split(".")[0]

        if self.model_provider == SupportModelProviders.Anthropic and self.model_api_version is None:
            self.model_api_version = "bedrock-2023-05-31"

        self.model_provider = AwsBedrockModelFactory.get_bedrock_model(
            model_provider=self.model_provider,
            model_api_version=self.model_api_version,
            model_id=self.model,
            aws_access_key=self.aws_access_key,
            aws_secret_key=self.aws_secret_key,
            region_name=self.region_name
        )

    # saas/proprietary
    def get_meta(self):
        return [{
            "model_deploy_type": "saas",
            "backend": "saas",
        }]
    
    async def async_get_meta(self):
        return await asyncfy_with_semaphore(self.get_meta)()
    
    async def async_stream_chat(
            self,
            tokenizer,
            ins: str,
            his: List[Tuple[str, str]] = [],
            max_length: int = 4096,
            top_p: float = 0.7,
            temperature: float = 0.9,
            **kwargs
    ):
        return await asyncfy_with_semaphore(self.stream_chat)(tokenizer, ins, his, max_length, top_p, temperature, **kwargs)

    def stream_chat(
            self,
            tokenizer,
            ins: str,
            his: List[Tuple[str, str]] = [],
            max_length: int = 4096,
            top_p: float = 0.7,
            temperature: float = 0.9,
            **kwargs
    ):
        try:
            return self.model_provider.generate(
                ins=ins,
                his=his,
                max_length=max_length,
                top_p=top_p,
                temperature=temperature,
                **kwargs
            )
        except Exception as e:
            logger.error(f"request aws bedrock failed: {e}")
        return None


class AWSBedrockModel(ABC):
    bedrock_client = None
    model_id = None
    model_api_version = None

    def __init__(
            self,
            model_id: str,
            region_name: str = None,
            aws_access_key: str = None,
            aws_secret_key: str = None,
            model_api_version: Optional[str] = None,

    ):
        self.model_id = model_id
        self.model_api_version = model_api_version
        self.bedrock = boto3.client(
            service_name='bedrock-runtime',
            aws_access_key_id=aws_access_key,
            aws_secret_access_key=aws_secret_key,
            region_name=region_name,
        )

    def generate(
            self,
            ins: str,
            his: List[Tuple[str, str]] = [],
            max_length: int = 4096,
            top_p: float = 0.7,
            temperature: float = 0.9,
            **kwargs):
        pass


class AnthropicModel(AWSBedrockModel):

    def generate(
            self,
            ins: str,
            his: List[Tuple[str, str]] = [],
            max_length: int = 4096,
            top_p: float = 0.7,
            temperature: float = 0.9,
            **kwargs
    ):
        body = self._generate_anthropic_request_body(ins, his)

        logger.info(f"【Byzer --> AWS-Bedrock({self.model_id})】:\n{body}")

        answer = None
        input_tokens = 0
        output_tokens = 0
        time_taken = 0
        start_time = time.monotonic()

        try:
            api_res = self.bedrock.invoke_model(
                body=json.dumps(body),
                modelId=self.model_id,
                accept='application/json',
                contentType='application/json'
            )
            time_taken = time.monotonic() - start_time
            response_body = json.loads(api_res.get('body').read())
            logger.info(
                f"【AWS-Bedrock({self.model_id}) --> Byzer】:\n{response_body}"
            )
            answer = response_body['content'][0]['text']
            input_tokens = response_body['usage']['input_tokens']
            output_tokens = response_body['usage']['output_tokens']
        except Exception as e:
            logger.error(f"request aws bedrock failed: {e}")
            answer = f"Exception occurred during the request, please try again: {e}" if not answer else answer

        return [(
            answer,
            {
                "metadata": {
                    "request_id": "",
                    "input_tokens_count": input_tokens,
                    "generated_tokens_count": output_tokens,
                    "time_cost": time_taken,
                    "first_token_time": -1.0,
                    "speed": float(output_tokens) / time_taken * 1000 if time_taken > 0 else 0,
                }
            }
        )]

    def _generate_anthropic_request_body(
            self,
            ins: str,
            his: List[Tuple[str, str]] = [],
            max_length: int = 4096,
            temperature: float = 0.9,
    ):
        messages = []
        for item in his:
            role, content = item["role"], item["content"]
            if role == "system":
                messages.append({"role": "user", "content": [{"type": "text", "text": content}]})
                messages.append({"role": "assistant", "content": [{"type": "text", "text": "OK"}]})
                continue
            messages.append({"role": role, "content": [{"type": "text", "text": content}]})

        if ins and len(messages) == 0:
            messages.append({"role": "user", "content": [{"type": "text", "text": ins}]})

        if ins and len(messages) > 0:
            last_message_role = messages[-1]["role"]
            if last_message_role == "user":
                messages.append({"role": "assistant", "content": [{"type": "text", "text": "OK"}]})
            messages.append({"role": "user", "content": [{"type": "text", "text": ins}]})

        anthropic_request_body = {
            "anthropic_version": self.model_api_version,
            "temperature": temperature,
            "max_tokens": max_length,
            "messages": messages
        }

        return anthropic_request_body


class MetaModel(AWSBedrockModel):

    def generate(
            self,
            ins: str,
            his: List[Tuple[str, str]] = [],
            max_length: int = 4096,
            top_p: float = 0.7,
            temperature: float = 0.9,
            **kwargs
    ):
        request_id = random_uuid() if "request_id" not in kwargs else kwargs["request_id"]

        fin_ins = self._generate_meta_instruction_from_history(ins, his)

        max_length = min(2048, max_length)

        meta_request_body = json.dumps({
            "prompt": fin_ins,
            "max_gen_len": max_length,
            "temperature": temperature,
            "top_p": top_p,
        })

        logger.info(f"Receiving request model: {self.model_id} messages: {meta_request_body}")

        answer = None
        input_tokens = 0
        output_tokens = 0
        time_taken = 0
        start_time = time.monotonic()

        try:
            api_res = self.bedrock.invoke_model(
                body=meta_request_body,
                modelId=self.model_id,
                accept='application/json',
                contentType='application/json'
            )
            time_taken = time.monotonic() - start_time
            response_body = json.loads(api_res.get('body').read())
            logger.info(
                f"Completed request {request_id}: "
                f"model: {self.model_id} "
                f"cost: {time_taken} "
                f"result: {response_body}"
            )
            answer = response_body['generation']
            input_tokens = response_body['prompt_token_count']
            output_tokens = response_body['generation_token_count']
        except Exception as e:
            logger.error(f"request aws bedrock failed: {e}")
            answer = f"Exception occurred during the request, please try again: {e}" if not answer else answer

        return [(
            answer,
            {
                "metadata": {
                    "request_id": "",
                    "input_tokens_count": input_tokens,
                    "generated_tokens_count": output_tokens,
                    "time_cost": time_taken,
                    "first_token_time": -1.0,
                    "speed": float(output_tokens) / time_taken * 1000 if time_taken > 0 else 0,
                }
            }
        )]

    def _generate_meta_instruction_from_history(
            self,
            ins: str,
            his: List[Dict[str, str]],
            role_mapping: Dict[str, str] = {
                "user": "Human",
                "assistant": "Assistant",
            }
    ):
        # if len(his) == 1 and ins == "":
        #     return his[0]['content'].strip() + "\n"

        new_his = []
        for item in his:
            if item["role"] == "system":
                new_his.append(item["content"])
                continue
            if item["role"] == "user":
                new_his.append(
                    f"[INST]{role_mapping[item['role']]}: {item['content']}")
                continue
            if item["role"] == "assistant":
                new_his.append(
                    f"{role_mapping[item['role']]}: {item['content']}[/INST]")
                continue

        # here we should make sure the user build the conversation string manually also
        # works. This means if the user do not provide  the history, then
        # we should treat ins as conversation string which the user build manually
        if len(new_his) > 0 and ins != "":
            new_his.append(f"[INST]{role_mapping['user']}: {ins}")
            new_his.append(f"{role_mapping['assistant']}:[/INST]")

        if len(new_his) > 0 and ins == "":
            new_his.append(f"{role_mapping['assistant']}[/INST]:")

        if len(new_his) == 0:
            new_his.append(ins)

        fin_ins = "\n".join(new_his)
        return fin_ins


class AwsBedrockModelFactory:
    @staticmethod
    def get_bedrock_model(
            model_provider: str,
            model_id: str,
            region_name: str = None,
            aws_access_key: str = None,
            aws_secret_key: str = None,
            model_api_version: Optional[str] = None,
    ) -> AWSBedrockModel:
        model_provider = model_provider.lower()
        logger.info(f"Using model_provider: {model_provider}")
        if SupportModelProviders.Anthropic == model_provider:
            return AnthropicModel(
                model_id=model_id,
                model_api_version=model_api_version,
                region_name=region_name,
                aws_access_key=aws_access_key,
                aws_secret_key=aws_secret_key
            )
        elif SupportModelProviders.Meta == model_provider:
            return MetaModel(
                model_id=model_id,
                model_api_version=model_api_version,
                region_name=region_name,
                aws_access_key=aws_access_key,
                aws_secret_key=aws_secret_key
            )
        else:
            raise ValueError(f"Unsupported model_provider: {model_provider}")



import glob
import os
from typing import *

import torch
from safetensors.torch import load_file
from transformers import CLIPTextModel

from byzerllm.stable_diffusion.api.events import event_handler
from byzerllm.stable_diffusion.api.events.generation import (
    LoadResourceEvent,
    PromptTokenizingEvent,
)
from byzerllm.stable_diffusion.config import stableDiffusionConfig

token_replaces = {}
loaded_embeddings = []


@event_handler()
def on_load_resource(e: LoadResourceEvent):
    global token_replaces, loaded_embeddings
    if not isinstance(e.pipe.text_encoder, CLIPTextModel):
        return
    embeddings_dir = os.path.join(stableDiffusionConfig.get_temp_dir(), "embeddings")
    embeddings = []
    for file in glob.glob(os.path.join(embeddings_dir, "**", "*"), recursive=True):
        safetensors = file.endswith(".safetensors")
        pt = file.endswith(".ckpt") or file.endswith(".pt")
        if safetensors or pt:
            embeddings.append(((pt, safetensors), file))

    if len(embeddings) == len(loaded_embeddings):
        if all(
            [
                embedding in loaded_embeddings
                for embedding in [embedding for _, embedding in embeddings]
            ]
        ):
            return

    token_replaces = {}
    loaded_embeddings = []

    for (pt, safetensors), file in embeddings:
        if safetensors:
            state_dict = load_file(file)
        else:
            state_dict = torch.load(file, map_location="cpu")

        if isinstance(state_dict, torch.Tensor):
            embedding = state_dict
        elif len(state_dict) == 1:
            embedding = next(iter(state_dict.values()))
        elif "string_to_param" in state_dict:
            embedding = state_dict["string_to_param"]["*"]

        token = os.path.splitext(os.path.basename(file))[0]

        embedding = embedding.to(
            dtype=e.pipe.text_encoder.dtype, device=e.pipe.text_encoder.device
        )

        is_multi_vector = len(embedding.shape) > 1 and embedding.shape[0] > 1

        if is_multi_vector:
            tokens = [token] + [f"{token}_{i}" for i in range(1, embedding.shape[0])]
            embeds = [e for e in embedding]  # noqa: C416
        else:
            tokens = [token]
            embeds = [embedding[0]] if len(embedding.shape) > 1 else [embedding]

        e.pipe.tokenizer.add_tokens(tokens)
        token_ids = e.pipe.tokenizer.convert_tokens_to_ids(tokens)

        e.pipe.text_encoder.resize_token_embeddings(len(e.pipe.tokenizer))
        for token_id, embedding in zip(token_ids, embeds):
            weight = e.pipe.text_encoder.get_input_embeddings().weight
            if weight.size()[1] != embedding.size()[0]:
                continue
            weight.data[token_id] = embedding

        loaded_embeddings.append(file)
        token_replaces[token_ids[0]] = token_ids


@event_handler()
def on_prompt_tokenizing(e: PromptTokenizingEvent):
    for token in e.text_tokens:
        if token in token_replaces:
            i = e.text_tokens.index(token)
            e.text_tokens[i : i + 1] = token_replaces[token]


def init():
    pass



import argparse
import os
import ray
import shlex
import jinja2
import yaml
from byzerllm.utils.client import ByzerLLM, InferBackend
from byzerllm.utils.client.types import Templates
from byzerllm.apps.byzer_storage.command import StorageSubCommand
from byzerllm.utils.client.entrypoints.openai.serve import serve, ServerArgs
import byzerllm
from byzerllm.lang import locales, lang 
from byzerllm.command_args import StoreNestedDict, get_command_args

def main():
    args = get_command_args()

    if args.file:
        with open(args.file, "r") as f:
            config = yaml.safe_load(f)
            for key, value in config.items():
                if key != "file":  # 排除 --file 参数本身   
                    ## key: ENV {{VARIABLE_NAME}}
                    if isinstance(value, str) and value.startswith("ENV"):  
                        template = jinja2.Template(value.removeprefix("ENV").strip())                  
                        value = template.render(os.environ)                        
                    setattr(args, key, value)
    
    print("Command Line Arguments:")
    print("-" * 50)
    for arg, value in vars(args).items():        
        if arg == "infer_params" and isinstance(value, dict) and "saas.api_key" in value:  
            import copy
            new_value = copy.deepcopy(value)
            new_value["saas.api_key"] = "******"
            print(f"{arg:20}: {new_value}")
        else:     
            print(f"{arg:20}: {value}")
    print("-" * 50)    

    if args.command == 'deploy':
        byzerllm.connect_cluster(address=args.ray_address)
        llm = ByzerLLM()
        if llm.is_model_exist(args.model):
            print(locales["already_deployed"][lang].format(args.model))
            return
        
        llm.setup_gpus_per_worker(args.gpus_per_worker).setup_cpus_per_worker(args.cpus_per_worker).setup_num_workers(args.num_workers)
        llm.setup_worker_concurrency(args.worker_concurrency)
        if not args.pretrained_model_type.startswith("saas"):
            if args.infer_backend == "vllm":
                llm.setup_infer_backend(InferBackend.VLLM)
            elif args.infer_backend == "llama_cpp": 
                llm.setup_infer_backend(InferBackend.LLAMA_CPP)
            elif args.infer_backend == "transformers":
                llm.setup_infer_backend(InferBackend.Transformers)
            elif args.infer_backend == "deepspeed":    
                llm.setup_infer_backend(InferBackend.DeepSpeed)
            else:
                raise ValueError("Invalid infer_backend")
        
        llm.deploy(model_path=args.model_path,
                pretrained_model_type=args.pretrained_model_type,
                udf_name=args.model,
                infer_params=args.infer_params or {})

        print(locales["deploy_success"][lang].format(args.model))

    elif args.command == 'query':
        byzerllm.connect_cluster(address=args.ray_address)

        llm_client = ByzerLLM()        
        if args.template == "default":
            llm_client.setup_template(args.model, template=Templates.default())
        elif args.template == "llama":
            llm_client.setup_template(args.model, template=Templates.llama())    
        elif args.template == "qwen":
            llm_client.setup_template(args.model, template=Templates.qwen())        
        elif args.template == "yi":
            llm_client.setup_template(args.model, template=Templates.yi())  
        elif args.template == "empty":
            llm_client.setup_template(args.model, template=Templates.empty())          
        else:
            llm_client.setup_template(args.model, args.template)

        resp = llm_client.chat_oai(model=args.model, conversations=[{
            "role": "user",
            "content": args.query,
        }])
        
        if args.output_file:
            ext = os.path.splitext(args.output_file)[1]
            if ext == ".mp3" or ext == ".wav":
                import base64
                with open(args.output_file, "wb") as f:
                    f.write(base64.b64decode(resp[0].output))
            else:        
                with open(args.output_file, "w") as f:
                    f.write(resp[0].output)
        else:        
            print(resp[0].output,flush=True)

    elif args.command == 'undeploy':
        byzerllm.connect_cluster(address=args.ray_address)

        llm = ByzerLLM()
        llm.undeploy(args.model,force=args.force)

        print(locales["undeploy_success"][lang].format(args.model))

    elif args.command == 'stat':
        import json
        byzerllm.connect_cluster(address=args.ray_address)
        model = ray.get_actor(args.model)
        v = ray.get(model.stat.remote())
        o = json.dumps(v, indent=4,ensure_ascii=False)
        print(o)    
        
    elif args.command == 'storage':
        if args.storage_command == "install":
            StorageSubCommand.install(args)
        elif args.storage_command == "start":
            StorageSubCommand.start(args) 
        elif args.storage_command == "stop":
            StorageSubCommand.stop(args)
        elif args.storage_command == "export":
            StorageSubCommand.export(args)
        elif args.storage_command == "collection":
            StorageSubCommand.collection(args)    
    
    elif args.command == 'serve':
        byzerllm.connect_cluster(address=args.ray_address) 
        llm_client = ByzerLLM()
        if args.served_model_name:
            llm_client.setup_template(args.served_model_name, args.template)                 
        server_args = ServerArgs(**{arg: getattr(args, arg) for arg in vars(ServerArgs())})                      
        serve(llm=llm_client, args=server_args)            


""" PyTorch Moss model."""

from typing import Optional, Tuple, Union

import torch
import torch.utils.checkpoint
from torch import nn
from torch.nn import CrossEntropyLoss
import transformers
from transformers.activations import ACT2FN
from transformers.modeling_utils import PreTrainedModel
from transformers.modeling_outputs import BaseModelOutputWithPast, CausalLMOutputWithPast
from transformers.utils import (
    add_code_sample_docstrings, 
    add_start_docstrings, 
    add_start_docstrings_to_model_forward, 
    logging
)

from .configuration_moss import MossConfig

logger = logging.get_logger(__name__)

_CHECKPOINT_FOR_DOC = "fnlp/moss-moon-003-base"
_CONFIG_FOR_DOC = "MossConfig"


MOSS_PRETRAINED_MODEL_ARCHIVE_LIST = [
    "fnlp/moss-moon-003-base",
    "fnlp/moss-moon-003-sft",
    "fnlp/moss-moon-003-sft-plugin",
    "fnlp/moss-moon-003-sft-int4",
    "fnlp/moss-moon-003-sft-plugin-int4",
    "fnlp/moss-moon-003-sft-int8",
    "fnlp/moss-moon-003-sft-plugin-int8",
]


# Copied from transformers.models.gptj.modeling_gptj.create_sinusoidal_positions
def create_sinusoidal_positions(num_pos: int, dim: int) -> torch.Tensor:
    inv_freq = 1.0 / (10000 ** (torch.arange(0, dim, 2) / dim))
    sinusoid_inp = torch.einsum("i , j -> i j", torch.arange(num_pos, dtype=torch.float), inv_freq).float()
    return torch.cat((torch.sin(sinusoid_inp), torch.cos(sinusoid_inp)), dim=1)


# Copied from transformers.models.gptj.modeling_gptj.rotate_every_two
def rotate_every_two(x: torch.Tensor) -> torch.Tensor:
    x1 = x[:, :, :, ::2]
    x2 = x[:, :, :, 1::2]
    x = torch.stack((-x2, x1), dim=-1)
    return x.flatten(-2)  # in einsum notation: rearrange(x, '... d j -> ... (d j)')


# Copied from transformers.models.gptj.modeling_gptj.apply_rotary_pos_emb
def apply_rotary_pos_emb(tensor: torch.Tensor, sin: torch.Tensor, cos: torch.Tensor) -> torch.Tensor:
    sin = torch.repeat_interleave(sin[:, :, None, :], 2, 3)
    cos = torch.repeat_interleave(cos[:, :, None, :], 2, 3)
    return (tensor * cos) + (rotate_every_two(tensor) * sin)


class MossAttention(nn.Module):
    def __init__(self, config):
        super().__init__()

        max_positions = config.max_position_embeddings
        self.register_buffer(
            "causal_mask",
            torch.tril(torch.ones((max_positions, max_positions), dtype=torch.bool)).view(
                1, 1, max_positions, max_positions
            ),
        )

        self.attn_dropout = nn.Dropout(config.attn_pdrop)
        self.resid_dropout = nn.Dropout(config.resid_pdrop)

        self.embed_dim = config.hidden_size
        self.num_attention_heads = config.num_attention_heads
        self.head_dim = self.embed_dim // self.num_attention_heads
        if self.head_dim * self.num_attention_heads != self.embed_dim:
            raise ValueError(
                f"embed_dim must be divisible by num_attention_heads (got `embed_dim`: {self.embed_dim} and"
                f" `num_attention_heads`: {self.num_attention_heads})."
            )
        self.scale_attn = torch.sqrt(torch.tensor(self.head_dim, dtype=torch.float32)).to(torch.get_default_dtype())
        self.qkv_proj = nn.Linear(self.embed_dim, self.embed_dim * 3, bias=False)

        self.out_proj = nn.Linear(self.embed_dim, self.embed_dim, bias=False)
        self.rotary_dim = config.rotary_dim
        pos_embd_dim = self.rotary_dim or self.embed_dim
        self.embed_positions = create_sinusoidal_positions(max_positions, pos_embd_dim)

    def _split_heads(self, x, n_head, dim_head, mp_num):
        reshaped = x.reshape(x.shape[:-1] + (n_head // mp_num, dim_head))
        reshaped = reshaped.reshape(x.shape[:-2] + (-1,) + reshaped.shape[-1:])
        return reshaped

    def _merge_heads(self, tensor, num_attention_heads, attn_head_size):
        """
        Merges attn_head_size dim and num_attn_heads dim into n_ctx
        """
        if len(tensor.shape) == 5:
            tensor = tensor.permute(0, 1, 3, 2, 4).contiguous()
        elif len(tensor.shape) == 4:
            tensor = tensor.permute(0, 2, 1, 3).contiguous()
        else:
            raise ValueError(f"Input tensor rank should be one of [4, 5], but is: {len(tensor.shape)}")
        new_shape = tensor.size()[:-2] + (num_attention_heads * attn_head_size,)
        return tensor.view(new_shape)

    def _attn(
        self,
        query,
        key,
        value,
        attention_mask=None,
        head_mask=None,
    ):
        # compute causal mask from causal mask buffer
        query_length, key_length = query.size(-2), key.size(-2)
        causal_mask = self.causal_mask[:, :, key_length - query_length : key_length, :key_length]

        # Keep the attention weights computation in fp32 to avoid overflow issues
        query = query.to(torch.float32)
        key = key.to(torch.float32)

        attn_weights = torch.matmul(query, key.transpose(-1, -2))

        attn_weights = attn_weights / self.scale_attn
        mask_value = torch.finfo(attn_weights.dtype).min
        # Need to be a tensor, otherwise we get error: `RuntimeError: expected scalar type float but found double`.
        # Need to be on the same device, otherwise `RuntimeError: ..., x and y to be on the same device`
        mask_value = torch.tensor(mask_value, dtype=attn_weights.dtype).to(attn_weights.device)
        attn_weights = torch.where(causal_mask, attn_weights, mask_value)

        if attention_mask is not None:
            # Apply the attention mask
            attn_weights = attn_weights + attention_mask

        attn_weights = nn.Softmax(dim=-1)(attn_weights)
        attn_weights = attn_weights.to(value.dtype)
        attn_weights = self.attn_dropout(attn_weights)

        # Mask heads if we want to
        if head_mask is not None:
            attn_weights = attn_weights * head_mask

        attn_output = torch.matmul(attn_weights, value)

        return attn_output, attn_weights

    def forward(
        self,
        hidden_states: Optional[torch.FloatTensor],
        layer_past: Optional[Tuple[torch.Tensor]] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        use_cache: Optional[bool] = False,
        output_attentions: Optional[bool] = False,
    ) -> Union[
        Tuple[torch.Tensor, Tuple[torch.Tensor]],
        Optional[Tuple[torch.Tensor, Tuple[torch.Tensor], Tuple[torch.Tensor, ...]]],
    ]:
        qkv = self.qkv_proj(hidden_states)
        # TODO(enijkamp): factor out number of logical TPU-v4 cores or make forward pass agnostic
        mp_num = 4
        qkv_split = qkv.reshape(qkv.shape[:-1] + (mp_num, -1))

        local_dim = self.head_dim * self.num_attention_heads // mp_num
        query, value, key = torch.split(qkv_split, local_dim, dim=-1)
        query = self._split_heads(query, self.num_attention_heads, self.head_dim, mp_num=mp_num)
        key = self._split_heads(key, self.num_attention_heads, self.head_dim, mp_num=mp_num)

        value = self._split_heads(value, self.num_attention_heads, self.head_dim, mp_num=mp_num)
        value = value.permute(0, 2, 1, 3)

        embed_positions = self.embed_positions
        if embed_positions.device != position_ids.device:
            embed_positions = embed_positions.to(position_ids.device)
            self.embed_positions = embed_positions

        sincos = embed_positions[position_ids]
        sin, cos = torch.split(sincos, sincos.shape[-1] // 2, dim=-1)

        if self.rotary_dim is not None:
            k_rot = key[:, :, :, : self.rotary_dim]
            k_pass = key[:, :, :, self.rotary_dim :]

            q_rot = query[:, :, :, : self.rotary_dim]
            q_pass = query[:, :, :, self.rotary_dim :]

            k_rot = apply_rotary_pos_emb(k_rot, sin, cos)
            q_rot = apply_rotary_pos_emb(q_rot, sin, cos)

            key = torch.cat([k_rot, k_pass], dim=-1)
            query = torch.cat([q_rot, q_pass], dim=-1)
        else:
            key = apply_rotary_pos_emb(key, sin, cos)
            query = apply_rotary_pos_emb(query, sin, cos)

        key = key.permute(0, 2, 1, 3)
        query = query.permute(0, 2, 1, 3)

        if layer_past is not None:
            past_key = layer_past[0]
            past_value = layer_past[1]
            key = torch.cat((past_key, key), dim=-2)
            value = torch.cat((past_value, value), dim=-2)

        if use_cache is True:
            present = (key, value)
        else:
            present = None

        # compute self-attention: V x Softmax(QK^T)
        attn_output, attn_weights = self._attn(query, key, value, attention_mask, head_mask)

        attn_output = self._merge_heads(attn_output, self.num_attention_heads, self.head_dim)
        attn_output = self.out_proj(attn_output)
        attn_output = self.resid_dropout(attn_output)

        outputs = (attn_output, present)
        if output_attentions:
            outputs += (attn_weights,)

        return outputs  # a, present, (attentions)


# Copied from transformers.models.gptj.modeling_gptj.GPTJMLP with GPTJ->Moss
class MossMLP(nn.Module):
    def __init__(self, intermediate_size, config):  # in MLP: intermediate_size= 4 * embed_dim
        super().__init__()
        embed_dim = config.n_embd

        self.fc_in = nn.Linear(embed_dim, intermediate_size)
        self.fc_out = nn.Linear(intermediate_size, embed_dim)

        self.act = ACT2FN[config.activation_function]
        self.dropout = nn.Dropout(config.resid_pdrop)

    def forward(self, hidden_states: Optional[torch.FloatTensor]) -> torch.FloatTensor:
        hidden_states = self.fc_in(hidden_states)
        hidden_states = self.act(hidden_states)
        hidden_states = self.fc_out(hidden_states)
        hidden_states = self.dropout(hidden_states)
        return hidden_states


# Copied from transformers.models.gptj.modeling_gptj.GPTJBlock with GPTJ->Moss
class MossBlock(nn.Module):
    def __init__(self, config):
        super().__init__()
        inner_dim = config.n_inner if config.n_inner is not None else 4 * config.n_embd
        self.ln_1 = nn.LayerNorm(config.n_embd, eps=config.layer_norm_epsilon)
        self.attn = MossAttention(config)
        self.mlp = MossMLP(inner_dim, config)

    def forward(
        self,
        hidden_states: Optional[torch.FloatTensor],
        layer_past: Optional[Tuple[torch.Tensor]] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        use_cache: Optional[bool] = False,
        output_attentions: Optional[bool] = False,
    ) -> Union[Tuple[torch.Tensor], Optional[Tuple[torch.Tensor, Tuple[torch.FloatTensor, ...]]]]:
        residual = hidden_states
        hidden_states = self.ln_1(hidden_states)
        attn_outputs = self.attn(
            hidden_states=hidden_states,
            layer_past=layer_past,
            attention_mask=attention_mask,
            position_ids=position_ids,
            head_mask=head_mask,
            use_cache=use_cache,
            output_attentions=output_attentions,
        )
        attn_output = attn_outputs[0]  # output_attn: a, present, (attentions)
        outputs = attn_outputs[1:]

        feed_forward_hidden_states = self.mlp(hidden_states)
        hidden_states = attn_output + feed_forward_hidden_states + residual

        if use_cache:
            outputs = (hidden_states,) + outputs
        else:
            outputs = (hidden_states,) + outputs[1:]

        return outputs  # hidden_states, present, (attentions)


class MossPreTrainedModel(PreTrainedModel):
    """
    An abstract class to handle weights initialization and a simple interface for downloading and loading pretrained
    models.
    """

    config_class = MossConfig
    base_model_prefix = "transformer"
    supports_gradient_checkpointing = True
    _no_split_modules = ["MossBlock"]

    def __init__(self, *inputs, **kwargs):
        super().__init__(*inputs, **kwargs)

    def _init_weights(self, module):
        """Initialize the weights."""
        if isinstance(module, (nn.Linear,)):
            # Slightly different from Mesh Transformer JAX which uses truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def _set_gradient_checkpointing(self, module, value=False):
        if isinstance(module, MossModel):
            module.gradient_checkpointing = value


MOSS_START_DOCSTRING = r"""
    This model is a PyTorch [torch.nn.Module](https://pytorch.org/docs/stable/nn.html#torch.nn.Module) sub-class. Use
    it as a regular PyTorch Module and refer to the PyTorch documentation for all matter related to general usage and
    behavior.

    Parameters:
        config ([`MossConfig`]): Model configuration class with all the parameters of the model.
            Initializing with a config file does not load the weights associated with the model, only the
            configuration. Check out the [`~PreTrainedModel.from_pretrained`] method to load the model weights.
"""

MOSS_INPUTS_DOCSTRING = r"""
    Args:
        input_ids (`torch.LongTensor` of shape `({0})`):
            Indices of input sequence tokens in the vocabulary.

            Indices can be obtained using [`AutoProcenizer`]. See [`PreTrainedTokenizer.encode`] and
            [`PreTrainedTokenizer.__call__`] for details.

            [What are input IDs?](../glossary#input-ids)
        attention_mask (`torch.FloatTensor` of shape `({0})`, *optional*):
            Mask to avoid performing attention on padding token indices. Mask values selected in `[0, 1]`:

            - 1 for tokens that are **not masked**,
            - 0 for tokens that are **masked**.

            [What are attention masks?](../glossary#attention-mask)
        token_type_ids (`torch.LongTensor` of shape `({0})`, *optional*):
            Segment token indices to indicate first and second portions of the inputs. Indices are selected in `[0,
            1]`:

            - 0 corresponds to a *sentence A* token,
            - 1 corresponds to a *sentence B* token.

            [What are token type IDs?](../glossary#token-type-ids)
        position_ids (`torch.LongTensor` of shape `({0})`, *optional*):
            Indices of positions of each input sequence tokens in the position embeddings. Selected in the range `[0,
            config.n_positions - 1]`.

            [What are position IDs?](../glossary#position-ids)
        head_mask (`torch.FloatTensor` of shape `(num_attention_heads,)` or `(n_layer, num_attention_heads)`, *optional*):
            Mask to nullify selected heads of the self-attention modules. Mask values selected in `[0, 1]`:

            - 1 indicates the head is **not masked**,
            - 0 indicates the head is **masked**.

        inputs_embeds (`torch.FloatTensor` of shape `({0}, hidden_dim)`, *optional*):
            Optionally, instead of passing `input_ids` you can choose to directly pass an embedded representation. This
            is useful if you want more control over how to convert *input_ids* indices into associated vectors than the
            model's internal embedding lookup matrix.
        output_attentions (`bool`, *optional*):
            Whether or not to return the attentions tensors of all attention layers. See `attentions` under returned
            tensors for more detail.
        output_hidden_states (`bool`, *optional*):
            Whether or not to return the hidden states of all layers. See `hidden_states` under returned tensors for
            more detail.
        return_dict (`bool`, *optional*):
            Whether or not to return a [`~utils.ModelOutput`] instead of a plain tuple.
"""


@add_start_docstrings(
    "The bare Moss Model transformer outputting raw hidden-states without any specific head on top.",
    MOSS_START_DOCSTRING,
)
class MossModel(MossPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)

        self.embed_dim = config.n_embd
        self.vocab_size = config.vocab_size
        self.wte = nn.Embedding(config.vocab_size, self.embed_dim)
        self.drop = nn.Dropout(config.embd_pdrop)
        self.h = nn.ModuleList([MossBlock(config) for _ in range(config.n_layer)])
        self.ln_f = nn.LayerNorm(self.embed_dim, eps=config.layer_norm_epsilon)
        self.rotary_dim = min(config.rotary_dim, config.n_ctx // config.num_attention_heads)

        self.gradient_checkpointing = False

        # Initialize weights and apply final processing
        self.post_init()

    def get_input_embeddings(self):
        return self.wte

    def set_input_embeddings(self, new_embeddings):
        self.wte = new_embeddings

    @add_start_docstrings_to_model_forward(MOSS_INPUTS_DOCSTRING.format("batch_size, sequence_length"))
    @add_code_sample_docstrings(
        checkpoint=_CHECKPOINT_FOR_DOC,
        output_type=BaseModelOutputWithPast,
        config_class=_CONFIG_FOR_DOC,
    )
    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[Tuple[Tuple[torch.Tensor]]] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        token_type_ids: Optional[torch.LongTensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple, BaseModelOutputWithPast]:
        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        use_cache = use_cache if use_cache is not None else self.config.use_cache
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if input_ids is not None and inputs_embeds is not None:
            raise ValueError("You cannot specify both input_ids and inputs_embeds at the same time")
        elif input_ids is not None:
            input_shape = input_ids.size()
            input_ids = input_ids.view(-1, input_shape[-1])
            batch_size = input_ids.shape[0]
        elif inputs_embeds is not None:
            input_shape = inputs_embeds.size()[:-1]
            batch_size = inputs_embeds.shape[0]
        else:
            raise ValueError("You have to specify either input_ids or inputs_embeds")

        device = input_ids.device if input_ids is not None else inputs_embeds.device

        if token_type_ids is not None:
            token_type_ids = token_type_ids.view(-1, input_shape[-1])

        if position_ids is not None:
            position_ids = position_ids.view(-1, input_shape[-1]).long()

        if past_key_values is None:
            past_length = 0
            past_key_values = tuple([None] * len(self.h))
        else:
            past_length = past_key_values[0][0].size(-2)

        if position_ids is None:
            position_ids = torch.arange(past_length, input_shape[-1] + past_length, dtype=torch.long, device=device)
            position_ids = position_ids.unsqueeze(0).view(-1, input_shape[-1])

        # Attention mask.
        if attention_mask is not None:
            if batch_size <= 0:
                raise ValueError("batch_size has to be defined and > 0")
            attention_mask = attention_mask.view(batch_size, -1)
            # We create a 3D attention mask from a 2D tensor mask.
            # Sizes are [batch_size, 1, 1, to_seq_length]
            # So we can broadcast to [batch_size, num_heads, from_seq_length, to_seq_length]
            # this attention mask is more simple than the triangular masking of causal attention
            # used in OpenAI GPT, we just need to prepare the broadcast dimension here.
            attention_mask = attention_mask[:, None, None, :]

            # Since attention_mask is 1.0 for positions we want to attend and 0.0 for
            # masked positions, this operation will create a tensor which is 0.0 for
            # positions we want to attend and the dtype's smallest value for masked positions.
            # Since we are adding it to the raw scores before the softmax, this is
            # effectively the same as removing these entirely.
            attention_mask = attention_mask.to(dtype=self.dtype)  # fp16 compatibility
            attention_mask = (1.0 - attention_mask) * torch.finfo(self.dtype).min

        # Prepare head mask if needed
        # 1.0 in head_mask indicate we keep the head
        # attention_probs has shape bsz x num_attention_heads x N x N
        # head_mask has shape n_layer x batch x num_attention_heads x N x N
        head_mask = self.get_head_mask(head_mask, self.config.n_layer)

        if inputs_embeds is None:
            inputs_embeds = self.wte(input_ids)

        hidden_states = inputs_embeds

        if token_type_ids is not None:
            token_type_embeds = self.wte(token_type_ids)
            hidden_states = hidden_states + token_type_embeds

        hidden_states = self.drop(hidden_states)

        output_shape = input_shape + (hidden_states.size(-1),)

        if self.gradient_checkpointing and self.training:
            if use_cache:
                logger.warning_once(
                    "`use_cache=True` is incompatible with `config.gradient_checkpointing=True`. Setting "
                    "`use_cache=False`..."
                )
                use_cache = False

        presents = () if use_cache else None
        all_self_attentions = () if output_attentions else None
        all_hidden_states = () if output_hidden_states else None
        for i, (block, layer_past) in enumerate(zip(self.h, past_key_values)):
            if output_hidden_states:
                all_hidden_states = all_hidden_states + (hidden_states,)

            if self.gradient_checkpointing and self.training:

                def create_custom_forward(module):
                    def custom_forward(*inputs):
                        # None for past_key_value
                        return module(*inputs, use_cache, output_attentions)

                    return custom_forward

                outputs = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(block),
                    hidden_states,
                    None,
                    attention_mask,
                    position_ids,
                    head_mask[i],
                )
            else:
                outputs = block(
                    hidden_states=hidden_states,
                    layer_past=layer_past,
                    attention_mask=attention_mask,
                    position_ids=position_ids,
                    head_mask=head_mask[i],
                    use_cache=use_cache,
                    output_attentions=output_attentions,
                )

            hidden_states = outputs[0]
            if use_cache is True:
                presents = presents + (outputs[1],)

            if output_attentions:
                all_self_attentions = all_self_attentions + (outputs[2 if use_cache else 1],)

        hidden_states = self.ln_f(hidden_states)

        hidden_states = hidden_states.view(output_shape)
        # Add last hidden state
        if output_hidden_states:
            all_hidden_states = all_hidden_states + (hidden_states,)

        if not return_dict:
            return tuple(v for v in [hidden_states, presents, all_hidden_states, all_self_attentions] if v is not None)

        return BaseModelOutputWithPast(
            last_hidden_state=hidden_states,
            past_key_values=presents,
            hidden_states=all_hidden_states,
            attentions=all_self_attentions,
        )


@add_start_docstrings(
    """
    The Moss Model transformer with a language modeling head on top.
    """,
    MOSS_START_DOCSTRING,
)
class MossForCausalLM(MossPreTrainedModel):
    _keys_to_ignore_on_load_missing = [r"h\.\d+\.attn\.causal_mask"]

    def __init__(self, config):
        super().__init__(config)
        if not hasattr(config, 'wbits'):
            config.wbits = 32
            config.groupsize = 128
            
        if config.wbits not in [4, 8, 32]:
            logger.warning(f'Specify `wbits` with 4, 8 or 32 to load the model. ')
        if config.wbits in [4, 8]:
            def noop(*args, **kwargs):
                pass
            torch.nn.init.kaiming_uniform_ = noop
            torch.nn.init.uniform_ = noop
            torch.nn.init.normal_ = noop

            torch.set_default_dtype(torch.half)
            transformers.modeling_utils._init_weights = False
            torch.set_default_dtype(torch.half)
        self.transformer = MossModel(config)
        self.lm_head = nn.Linear(config.n_embd, config.vocab_size)
        if config.wbits in [4, 8]:
            torch.set_default_dtype(torch.float)
            transformers.modeling_utils._init_weights = True
            self.quantize(config.wbits, config.groupsize)
        # Initialize weights and apply final processing
        self.post_init()

    def get_output_embeddings(self):
        return self.lm_head

    def set_output_embeddings(self, new_embeddings):
        self.lm_head = new_embeddings

    def prepare_inputs_for_generation(self, input_ids, past_key_values=None, **kwargs):
        token_type_ids = kwargs.get("token_type_ids", None)
        # only last token for inputs_ids if past is defined in kwargs
        if past_key_values:
            input_ids = input_ids[:, -1].unsqueeze(-1)
            if token_type_ids is not None:
                token_type_ids = token_type_ids[:, -1].unsqueeze(-1)

        attention_mask = kwargs.get("attention_mask", None)
        position_ids = kwargs.get("position_ids", None)

        if attention_mask is not None and position_ids is None:
            # create position_ids on the fly for batch generation
            position_ids = attention_mask.long().cumsum(-1) - 1
            position_ids.masked_fill_(attention_mask == 0, 1)
            if past_key_values:
                position_ids = position_ids[:, -1].unsqueeze(-1)

        return {
            "input_ids": input_ids,
            "past_key_values": past_key_values,
            "use_cache": kwargs.get("use_cache"),
            "position_ids": position_ids,
            "attention_mask": attention_mask,
            "token_type_ids": token_type_ids,
        }

    @add_start_docstrings_to_model_forward(MOSS_INPUTS_DOCSTRING.format("batch_size, sequence_length"))
    @add_code_sample_docstrings(
        checkpoint=_CHECKPOINT_FOR_DOC,
        output_type=CausalLMOutputWithPast,
        config_class=_CONFIG_FOR_DOC,
    )
    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[Tuple[Tuple[torch.Tensor]]] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        token_type_ids: Optional[torch.LongTensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple, CausalLMOutputWithPast]:
        r"""
        labels (`torch.LongTensor` of shape `(batch_size, sequence_length)`, *optional*):
            Labels for language modeling. Note that the labels **are shifted** inside the model, i.e. you can set
            `labels = input_ids` Indices are selected in `[-100, 0, ..., config.vocab_size]` All labels set to `-100`
            are ignored (masked), the loss is only computed for labels in `[0, ..., config.vocab_size]`
        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        transformer_outputs = self.transformer(
            input_ids,
            past_key_values=past_key_values,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        hidden_states = transformer_outputs[0]

        # make sure sampling in fp16 works correctly and
        # compute loss in fp32 to match with mesh-tf version
        # https://github.com/EleutherAI/gpt-neo/blob/89ce74164da2fb16179106f54e2269b5da8db333/models/gpt2/gpt2.py#L179
        lm_logits = self.lm_head(hidden_states).to(torch.float32)

        loss = None
        if labels is not None:
            # Shift so that tokens < n predict n
            shift_logits = lm_logits[..., :-1, :].contiguous()
            shift_labels = labels[..., 1:].contiguous()
            # Flatten the tokens
            loss_fct = CrossEntropyLoss()
            loss = loss_fct(shift_logits.view(-1, shift_logits.size(-1)), shift_labels.view(-1))

            loss = loss.to(hidden_states.dtype)

        if not return_dict:
            output = (lm_logits,) + transformer_outputs[1:]
            return ((loss,) + output) if loss is not None else output

        return CausalLMOutputWithPast(
            loss=loss,
            logits=lm_logits,
            past_key_values=transformer_outputs.past_key_values,
            hidden_states=transformer_outputs.hidden_states,
            attentions=transformer_outputs.attentions,
        )

    @staticmethod
    def _reorder_cache(
        past_key_values: Tuple[Tuple[torch.Tensor]], beam_idx: torch.Tensor
    ) -> Tuple[Tuple[torch.Tensor]]:
        """
        This function is used to re-order the `past_key_values` cache if [`~PretrainedModel.beam_search`] or
        [`~PretrainedModel.beam_sample`] is called. This is required to match `past_key_values` with the correct
        beam_idx at every generation step.
        """
        return tuple(
            tuple(past_state.index_select(0, beam_idx.to(past_state.device)) for past_state in layer_past)
            for layer_past in past_key_values
        )

    def quantize(self, wbits, groupsize):
        from .quantization import quantize_with_gptq
        return quantize_with_gptq(self, wbits, groupsize)

